from collections import OrderedDict
import os
import sys

import numpy as np
import scipy.linalg
import scipy.signal

from toast import Operator
from toast.mpi import MPI

from ..timing import gather_timers, GlobalTimers
from toast.timing import function_timer, Timer
from toast.utils import Logger, Environment
from .sim_det_map import OpSimScan
from .todmap_math import OpAccumDiag, OpScanScale, OpScanMask
from ..tod import OpCacheClear, OpCacheCopy, OpCacheInit, OpFlagsApply, OpFlagGaps
from ..map import covariance_apply, covariance_invert, DistPixels, covariance_rcond
from .. import qarray as qa

from .._libtoast import add_offsets_to_signal, project_signal_offsets


XAXIS, YAXIS, ZAXIS = np.eye(3)

temporary_names = set()


def get_temporary_name():
    i = 0
    while True:
        name = "temporary{:03}".format(i)
        if name not in temporary_names:
            break
        i += 1
    temporary_names.add(name)
    return name


def free_temporary_name(name):
    temporary_names.remove(name)


class TOASTMatrix:
    def apply(self, vector, inplace=False):
        """Every TOASTMatrix can apply itself to a distributed vectors
        of signal, map or template offsets as is appropriate.
        """
        raise NotImplementedError("Virtual apply not implemented in derived class")

    def apply_transpose(self, vector, inplace=False):
        """Every TOASTMatrix can apply itself to a distributed vectors
        of signal, map or template offsets as is appropriate.
        """
        raise NotImplementedError(
            "Virtual apply_transpose not implemented in derived class"
        )


class TOASTVector:
    def dot(self, other):
        raise NotImplementedError("Virtual dot not implemented in derived class")


class UnitMatrix(TOASTMatrix):
    def apply(self, vector, inplace=False):
        if inplace:
            outvec = vector
        else:
            outvec = vector.copy()
        return outvec


class TODTemplate:
    """Parent class for all templates that can be registered with
    TemplateMatrix
    """

    name = None
    namplitude = 0
    comm = None

    def __init___(self, *args, **kwargs):
        raise NotImplementedError("Derived class must implement __init__()")

    def add_to_signal(self, signal, amplitudes):
        """signal += F.a"""
        raise NotImplementedError("Derived class must implement add_to_signal()")

    def project_signal(self, signal, amplitudes):
        """a += F^T.signal"""
        raise NotImplementedError("Derived class must implement project_signal()")

    def add_prior(self, amplitudes_in, amplitudes_out):
        """a' += C_a^{-1}.a"""
        # Not all TODTemplates implement the prior
        return

    def apply_precond(self, amplitudes_in, amplitudes_out):
        """a' = M^{-1}.a"""
        raise NotImplementedError("Derived class must implement apply_precond()")


class SubharmonicTemplate(TODTemplate):
    """This class represents sub-harmonic noise fluctuations.

    Sub-harmonic means that the characteristic frequency of the noise
    modes is lower than 1/T where T is the length of the interval
    being fitted.
    """

    name = "subharmonic"

    def __init__(
        self,
        data,
        detweights,
        order=1,
        intervals=None,
        common_flags=None,
        common_flag_mask=1,
        flags=None,
        flag_mask=1,
    ):
        self.data = data
        self.detweights = detweights
        self.order = order
        self.intervals = intervals
        self.common_flags = common_flags
        self.common_flag_mask = common_flag_mask
        self.flags = flags
        self.flag_mask = flag_mask
        self._last_nsamp = None
        self._last_templates = None
        self.get_steps_and_preconditioner()

    def get_steps_and_preconditioner(self):
        """Assign each template an amplitude"""
        self.templates = []
        self.slices = []
        self.preconditioners = []
        for iobs, obs in enumerate(self.data.obs):
            tod = obs["tod"]
            common_flags = tod.local_common_flags(self.common_flags)
            common_flags = (common_flags & self.common_flag_mask) != 0
            if (self.intervals is not None) and (self.intervals in obs):
                intervals = obs[self.intervals]
            else:
                intervals = None
            local_intervals = tod.local_intervals(intervals)
            slices = {}  # this observation
            preconditioners = {}  # this observation
            for ival in local_intervals:
                todslice = slice(ival.first, ival.last + 1)
                for idet, det in enumerate(tod.local_dets):
                    ind = slice(self.namplitude, self.namplitude + self.order + 1)
                    self.templates.append([ind, iobs, det, todslice])
                    self.namplitude += self.order + 1
                    preconditioner = self._get_preconditioner(
                        det, tod, todslice, common_flags, self.detweights[iobs][det]
                    )
                    if det not in preconditioners:
                        preconditioners[det] = []
                        slices[det] = []
                    preconditioners[det].append(preconditioner)
                    slices[det].append(ind)
            self.slices.append(slices)
            self.preconditioners.append(preconditioners)
        return

    def _get_preconditioner(self, det, tod, todslice, common_flags, detweight):
        """Calculate the preconditioner for the given interval and detector"""
        flags = tod.local_flags(det, self.flags)[todslice]
        good = (flags & self.flag_mask) == 0
        good[common_flags[todslice]] = False
        norder = self.order + 1
        preconditioner = np.zeros([norder, norder])
        templates = self._get_templates(todslice)
        for row in range(norder):
            for col in range(row, norder):
                preconditioner[row, col] = np.dot(
                    templates[row][good], templates[col][good]
                )
                preconditioner[row, col] *= detweight
                if row != col:
                    preconditioner[col, row] = preconditioner[row, col]
        preconditioner = np.linalg.inv(preconditioner)
        return preconditioner

    def add_to_signal(self, signal, amplitudes):
        subharmonic_amplitudes = amplitudes[self.name]
        for ibase, (ind, iobs, det, todslice) in enumerate(self.templates):
            templates = self._get_templates(todslice)
            amps = subharmonic_amplitudes[ind]
            for template, amplitude in zip(templates, amps):
                signal[iobs, det, todslice] += template * amplitude
        return

    def _get_templates(self, todslice):
        """Develop hierarchy of subharmonic modes matching the given length

        The basis functions are (orthogonal) Legendre polynomials
        """
        nsamp = todslice.stop - todslice.start
        if nsamp != self._last_nsamp:
            templates = np.zeros([self.order + 1, nsamp])
            r = np.linspace(-1, 1, nsamp)
            for order in range(self.order + 1):
                if order == 0:
                    templates[order] = 1
                elif order == 1:
                    templates[order] = r
                else:
                    templates[order] = (
                        (2 * order - 1) * r * templates[order - 1]
                        - (order - 1) * templates[order - 2]
                    ) / order
            self._last_nsamp = nsamp
            self._last_templates = templates
        return self._last_templates

    def project_signal(self, signal, amplitudes):
        subharmonic_amplitudes = amplitudes[self.name]
        for ibase, (ind, iobs, det, todslice) in enumerate(self.templates):
            templates = self._get_templates(todslice)
            amps = subharmonic_amplitudes[ind]
            for order, template in enumerate(templates):
                amps[order] = np.dot(signal[iobs, det, todslice], template)
        pass

    def apply_precond(self, amplitudes_in, amplitudes_out):
        """Standard diagonal preconditioner accounting for the fact that
        the templates are not orthogonal in the presence of flagging and masking
        """
        subharmonic_amplitudes_in = amplitudes_in[self.name]
        subharmonic_amplitudes_out = amplitudes_out[self.name]
        for iobs, obs in enumerate(self.data.obs):
            tod = obs["tod"]
            for det in tod.local_dets:
                slices = self.slices[iobs][det]
                preconditioners = self.preconditioners[iobs][det]
                for ind, preconditioner in zip(slices, preconditioners):
                    subharmonic_amplitudes_out[ind] = np.dot(
                        preconditioner, subharmonic_amplitudes_in[ind]
                    )
        return


class Fourier2DTemplate(TODTemplate):
    """This class represents atmospheric fluctuations in front of the
    focalplane as 2D Fourier modes."""

    name = "Fourier2D"

    def __init__(
        self,
        data,
        detweights,
        focalplane_radius=None,  # degrees
        order=1,
        fit_subharmonics=True,
        intervals=None,
        common_flags=None,
        common_flag_mask=1,
        flags=None,
        flag_mask=1,
        correlation_length=10,
        correlation_amplitude=10,
    ):
        self.data = data
        self.comm = data.comm.comm_group
        self.detweights = detweights
        self.focalplane_radius = focalplane_radius
        self.order = order
        self.fit_subharmonics = fit_subharmonics
        self.norder = order + 1
        self.nmode = (2 * order) ** 2 + 1
        if self.fit_subharmonics:
            self.nmode += 2
        self.intervals = intervals
        self.common_flags = common_flags
        self.common_flag_mask = common_flag_mask
        self.flags = flags
        self.flag_mask = flag_mask
        self._get_templates()
        self.correlation_length = correlation_length
        self.correlation_amplitude = correlation_amplitude
        if correlation_length:
            self._get_prior()
        return

    @function_timer
    def _get_prior(self):
        """Evaluate C_a^{-1} for the 2D polynomial coefficients based
        on the correlation length.
        """
        if self.correlation_length:
            # Correlation length is given in seconds and we cannot assume
            # that each observation has the same sampling rate.  Therefore,
            # we will build the filter for each observation
            self.filters = []  # all observations
            self.preconditioners = []  # all observations
            for iobs, obs in enumerate(self.data.obs):
                tod = obs["tod"]
                times = tod.local_times()
                corr = (
                    np.exp((times[0] - times) / self.correlation_length)
                    * self.correlation_amplitude
                )
                ihalf = times.size // 2
                corr[ihalf + 1 :] = corr[ihalf - 1 : 0 : -1]
                fcorr = np.fft.rfft(corr)
                invcorr = np.fft.irfft(1 / fcorr)
                self.filters.append(invcorr)
            # Scale the filter by the prescribed correlation strength
            # and the number of modes at each angular scale
            self.filter_scale = np.zeros(self.nmode)
            self.filter_scale[0] = 1
            offset = 1
            if self.fit_subharmonics:
                self.filter_scale[1:3] = 2
                offset += 2
            self.filter_scale[offset:] = 4
            self.filter_scale *= self.correlation_amplitude
        return

    @function_timer
    def _get_templates(self):
        """Evaluate and normalize the polynomial templates.

        Each template corresponds to a fixed value for each detector
        and depends on the position of the detector.
        """
        self.templates = []

        def evaluate_template(theta, phi, radius):
            values = np.zeros(self.nmode)
            values[0] = 1
            offset = 1
            if self.fit_subharmonics:
                values[1:3] = theta / radius, phi / radius
                offset += 2
            if self.order > 0:
                rinv = np.pi / radius
                orders = np.arange(self.order) + 1
                thetavec = np.zeros(self.order * 2)
                phivec = np.zeros(self.order * 2)
                thetavec[::2] = np.cos(orders * theta * rinv)
                thetavec[1::2] = np.sin(orders * theta * rinv)
                phivec[::2] = np.cos(orders * phi * rinv)
                phivec[1::2] = np.sin(orders * phi * rinv)
                values[offset:] = np.outer(thetavec, phivec).ravel()
            return values

        self.norms = []
        for iobs, obs in enumerate(self.data.obs):
            tod = obs["tod"]
            common_flags = tod.local_common_flags(self.common_flags)
            common_flags = (common_flags & self.common_flag_mask) != 0
            nsample = tod.total_samples
            obs_templates = {}
            focalplane = obs["focalplane"]
            if self.focalplane_radius:
                radius = np.radians(self.focalplane_radius)
            else:
                try:
                    radius = np.radians(focalplane.radius)
                except AttributeError:
                    # Focalplane is just a dictionary
                    radius = np.radians(obs["fpradius"])
            norms = np.zeros([nsample, self.nmode])
            local_offset, local_nsample = tod.local_samples
            todslice = slice(local_offset, local_offset + local_nsample)
            for det in tod.local_dets:
                flags = tod.local_flags(det, self.flags)
                good = ((flags & self.flag_mask) | common_flags) == 0
                detweight = self.detweights[iobs][det]
                det_quat = focalplane[det]["quat"]
                x, y, z = qa.rotate(det_quat, ZAXIS)
                theta, phi = np.arcsin([x, y])
                obs_templates[det] = evaluate_template(theta, phi, radius)
                norms[todslice] += np.outer(good, obs_templates[det] ** 2 * detweight)
            self.comm.allreduce(norms)
            good = norms != 0
            norms[good] = 1 / norms[good]
            self.norms.append(norms.ravel())
            self.templates.append(obs_templates)
            self.namplitude += nsample * self.nmode

        self.norms = np.hstack(self.norms)

        return

    @function_timer
    def add_to_signal(self, signal, amplitudes):
        """signal += F.a"""
        poly_amplitudes = amplitudes[self.name]
        amplitude_offset = 0
        for iobs, obs in enumerate(self.data.obs):
            tod = obs["tod"]
            nsample = tod.total_samples
            # For each observation, sample indices start from 0
            local_offset, local_nsample = tod.local_samples
            todslice = slice(local_offset, local_offset + local_nsample)
            obs_amplitudes = poly_amplitudes[
                amplitude_offset : amplitude_offset + nsample * self.nmode
            ].reshape([nsample, self.nmode])[todslice]
            for det in tod.local_dets:
                templates = self.templates[iobs][det]
                signal[iobs, det, todslice] += np.sum(obs_amplitudes * templates, 1)
            amplitude_offset += nsample * self.nmode
        return

    @function_timer
    def project_signal(self, signal, amplitudes):
        """a += F^T.signal"""
        poly_amplitudes = amplitudes[self.name]
        amplitude_offset = 0
        for iobs, obs in enumerate(self.data.obs):
            tod = obs["tod"]
            nsample = tod.total_samples
            # For each observation, sample indices start from 0
            local_offset, local_nsample = tod.local_samples
            todslice = slice(local_offset, local_offset + local_nsample)
            obs_amplitudes = poly_amplitudes[
                amplitude_offset : amplitude_offset + nsample * self.nmode
            ].reshape([nsample, self.nmode])
            if self.comm is not None:
                my_amplitudes = np.zeros_like(obs_amplitudes)
            else:
                my_amplitudes = obs_amplitudes
            for det in tod.local_dets:
                templates = self.templates[iobs][det]
                my_amplitudes[todslice] += np.outer(
                    signal[iobs, det, todslice], templates
                )
            if self.comm is not None:
                self.comm.allreduce(my_amplitudes)
                obs_amplitudes += my_amplitudes
            amplitude_offset += nsample * self.nmode
        return

    def add_prior(self, amplitudes_in, amplitudes_out):
        """a' += C_a^{-1}.a"""
        if self.correlation_length:
            poly_amplitudes_in = amplitudes_in[self.name]
            poly_amplitudes_out = amplitudes_out[self.name]
            amplitude_offset = 0
            for obs, noisefilter in zip(self.data.obs, self.filters):
                tod = obs["tod"]
                nsample = tod.total_samples
                obs_amplitudes_in = poly_amplitudes_in[
                    amplitude_offset : amplitude_offset + nsample * self.nmode
                ].reshape([nsample, self.nmode])
                obs_amplitudes_out = poly_amplitudes_out[
                    amplitude_offset : amplitude_offset + nsample * self.nmode
                ].reshape([nsample, self.nmode])
                # import pdb
                # import matplotlib.pyplot as plt
                # pdb.set_trace()
                for mode in range(self.nmode):
                    scale = self.filter_scale[mode]
                    obs_amplitudes_out[:, mode] += scipy.signal.convolve(
                        obs_amplitudes_in[:, mode],
                        noisefilter * scale,
                        mode="same",
                    )
                amplitude_offset += nsample * self.nmode
        return

    def apply_precond(self, amplitudes_in, amplitudes_out):
        """a' = M^{-1}.a"""
        poly_amplitudes_in = amplitudes_in[self.name]
        poly_amplitudes_out = amplitudes_out[self.name]
        poly_amplitudes_out[:] = poly_amplitudes_in * self.norms
        return


class OffsetTemplate(TODTemplate):
    """This class represents noise fluctuations as a step function"""

    name = "offset"

    def __init__(
        self,
        data,
        detweights,
        step_length=1000000,
        intervals=None,
        use_noise_prior=True,
        common_flags=None,
        common_flag_mask=1,
        flags=None,
        flag_mask=1,
        precond_width=20,
    ):
        self.data = data
        self.detweights = detweights
        self.step_length = step_length
        self.intervals = intervals
        self.common_flags = common_flags
        self.common_flag_mask = common_flag_mask
        self.flags = flags
        self.flag_mask = flag_mask
        self.precond_width = precond_width
        self.get_steps()
        self.use_noise_prior = use_noise_prior
        if self.use_noise_prior:
            self.get_filters_and_preconditioners()
        return

    @function_timer
    def get_filters_and_preconditioners(self):
        """Compute and store the filter and associated preconditioner
        for every detector and every observation
        """
        log = Logger.get()
        self.filters = []  # all observations
        self.preconditioners = []  # all observations
        for iobs, obs in enumerate(self.data.obs):
            if "noise" not in obs:
                # If the observations do not include noise PSD:s, we
                # we cannot build filters.
                if len(self.filters) > 0:
                    log.warning(
                        'Observation "{}" does not have noise information'
                        "".format(obs["name"])
                    )
                continue
            tod = obs["tod"]
            # Determine the binning for the noise prior
            times = tod.local_times()
            dtime = np.amin(np.diff(times))
            fsample = 1 / dtime
            obstime = times[-1] - times[0]
            tbase = self.step_length
            fbase = 1 / tbase
            powmin = np.floor(np.log10(1 / obstime)) - 1
            powmax = min(np.ceil(np.log10(1 / tbase)) + 2, fsample)
            freq = np.logspace(powmin, powmax, 1000)
            # Now build the filter for each detector
            noise = obs["noise"]
            noisefilters = {}  # this observation
            preconditioners = {}  # this observation
            for det in tod.local_dets:
                offset_psd = self._get_offset_psd(noise, freq, det)
                # Store real space filters for every interval and every detector.
                (
                    noisefilters[det],
                    preconditioners[det],
                ) = self._get_noisefilter_and_preconditioner(
                    freq, offset_psd, self.offset_slices[iobs][det]
                )
            self.filters.append(noisefilters)
            self.preconditioners.append(preconditioners)
        return

    @function_timer
    def _get_offset_psd(self, noise, freq, det):
        psdfreq = noise.freq(det)
        psd = noise.psd(det)
        rate = noise.rate(det)
        # Remove the white noise component from the PSD
        psd = psd.copy() * np.sqrt(rate)
        psd -= np.amin(psd[psdfreq > 1.0])
        psd[psd < 1e-30] = 1e-30

        # The calculation of `offset_psd` is from Keihänen, E. et al:
        # "Making CMB temperature and polarization maps with Madam",
        # A&A 510:A57, 2010
        logfreq = np.log(psdfreq)
        logpsd = np.log(psd)

        def interpolate_psd(x):
            result = np.zeros(x.size)
            good = np.abs(x) > 1e-10
            logx = np.log(np.abs(x[good]))
            logresult = np.interp(logx, logfreq, logpsd)
            result[good] = np.exp(logresult)
            return result

        def g(x):
            bad = np.abs(x) < 1e-10
            good = np.logical_not(bad)
            arg = np.pi * x[good]
            result = bad.astype(np.float64)
            result[good] = (np.sin(arg) / arg) ** 2
            return result

        tbase = self.step_length
        fbase = 1 / tbase
        offset_psd = interpolate_psd(freq) * g(freq * tbase)
        for m in range(1, 2):
            offset_psd += interpolate_psd(freq + m * fbase) * g(freq * tbase + m)
            offset_psd += interpolate_psd(freq - m * fbase) * g(freq * tbase - m)
        offset_psd *= fbase
        return offset_psd

    @function_timer
    def _get_noisefilter_and_preconditioner(self, freq, offset_psd, offset_slices):
        logfreq = np.log(freq)
        logpsd = np.log(offset_psd)
        logfilter = np.log(1 / offset_psd)

        def interpolate(x, psd):
            result = np.zeros(x.size)
            good = np.abs(x) > 1e-10
            logx = np.log(np.abs(x[good]))
            logresult = np.interp(logx, logfreq, psd)
            result[good] = np.exp(logresult)
            return result

        def truncate(noisefilter, lim=1e-4):
            icenter = noisefilter.size // 2
            ind = np.abs(noisefilter[:icenter]) > np.abs(noisefilter[0]) * lim
            icut = np.argwhere(ind)[-1][0]
            if icut % 2 == 0:
                icut += 1
            noisefilter = np.roll(noisefilter, icenter)
            noisefilter = noisefilter[icenter - icut : icenter + icut + 1]
            return noisefilter

        noisefilters = []
        preconditioners = []
        for offset_slice, sigmasqs in offset_slices:
            nstep = offset_slice.stop - offset_slice.start
            filterlen = nstep * 2 + 1
            filterfreq = np.fft.rfftfreq(filterlen, self.step_length)
            noisefilter = truncate(np.fft.irfft(interpolate(filterfreq, logfilter)))
            noisefilters.append(noisefilter)
            # Build the band-diagonal preconditioner
            if self.precond_width <= 1:
                # Compute C_a prior
                preconditioner = truncate(np.fft.irfft(interpolate(filterfreq, logpsd)))
            else:
                # Compute Cholesky decomposition prior
                wband = min(self.precond_width, noisefilter.size // 2)
                precond_width = max(wband, min(self.precond_width, nstep))
                icenter = noisefilter.size // 2
                preconditioner = np.zeros([precond_width, nstep], dtype=np.float64)
                preconditioner[0] = sigmasqs
                preconditioner[:wband, :] += np.repeat(
                    noisefilter[icenter : icenter + wband, np.newaxis], nstep, 1
                )
                lower = True
                scipy.linalg.cholesky_banded(
                    preconditioner, overwrite_ab=True, lower=lower, check_finite=True
                )
            preconditioners.append((preconditioner, lower))
        return noisefilters, preconditioners

    @function_timer
    def get_steps(self):
        """Divide each interval into offset steps"""
        self.offset_templates = []
        self.offset_slices = []  # slices in all observations
        for iobs, obs in enumerate(self.data.obs):
            tod = obs["tod"]
            common_flags = tod.local_common_flags(self.common_flags)
            common_flags = (common_flags & self.common_flag_mask) != 0
            if (self.intervals is not None) and (self.intervals in obs):
                intervals = obs[self.intervals]
            else:
                intervals = None
            local_intervals = tod.local_intervals(intervals)
            times = tod.local_times()
            offset_slices = {}  # slices in this observation
            for ival in local_intervals:
                length = times[ival.last] - times[ival.first]
                nbase = int(np.ceil(length / self.step_length))
                # Divide the interval into steps, allowing for irregular sampling
                todslices = []
                start_times = np.arange(nbase) * self.step_length + ival.start
                start_indices = np.searchsorted(times, start_times)
                stop_indices = np.hstack([start_indices[1:], [ival.last]])
                todslices = []
                for istart, istop in zip(start_indices, stop_indices):
                    todslices.append(slice(istart, istop))
                for idet, det in enumerate(tod.local_dets):
                    istart = self.namplitude
                    sigmasqs = []
                    for todslice in todslices:
                        sigmasq = self._get_sigmasq(
                            tod, det, todslice, common_flags, self.detweights[iobs][det]
                        )
                        # Register the baseline offset
                        self.offset_templates.append(
                            [self.namplitude, iobs, det, todslice, sigmasq]
                        )
                        sigmasqs.append(sigmasq)
                        self.namplitude += 1
                    # Keep a record of ranges of offsets that correspond
                    # to one detector and one interval.
                    # This is the domain we apply the noise filter in.
                    if det not in offset_slices:
                        offset_slices[det] = []
                    offset_slices[det].append(
                        (slice(istart, self.namplitude), sigmasqs)
                    )
            self.offset_slices.append(offset_slices)
        return

    @function_timer
    def _get_sigmasq(self, tod, det, todslice, common_flags, detweight):
        """calculate a rough estimate of the baseline variance
        for diagonal preconditioner
        """
        flags = tod.local_flags(det, self.flags)[todslice]
        good = (flags & self.flag_mask) == 0
        good[common_flags[todslice]] = False
        ngood = np.sum(good)
        sigmasq = 1
        if detweight != 0:
            sigmasq /= detweight
        if ngood != 0:
            sigmasq /= ngood
        return sigmasq

    @function_timer
    def add_to_signal(self, signal, amplitudes):
        offset_amplitudes = amplitudes[self.name]
        last_obs = None
        last_det = None
        last_ref = None
        todslices = []
        itemplates = []
        for itemplate, iobs, det, todslice, sigmasq in self.offset_templates:
            if iobs != last_obs or det != last_det:
                if len(todslices) != 0:
                    add_offsets_to_signal(
                        last_ref, todslices, offset_amplitudes, np.array(itemplates)
                    )
                todslices = []
                itemplates = []
                last_obs = iobs
                last_det = det
                last_ref = signal[iobs, det, :]
            todslices.append(todslice)
            itemplates.append(itemplate)
        if len(todslices) != 0:
            add_offsets_to_signal(
                last_ref, todslices, offset_amplitudes, np.array(itemplates)
            )
        return

    @function_timer
    def project_signal(self, signal, amplitudes):
        offset_amplitudes = amplitudes[self.name]
        last_obs = None
        last_det = None
        last_ref = None
        todslices = []
        itemplates = []
        for itemplate, iobs, det, todslice, sqsigma in self.offset_templates:
            if iobs != last_obs or det != last_det:
                if len(todslices) != 0:
                    project_signal_offsets(
                        last_ref, todslices, offset_amplitudes, np.array(itemplates)
                    )
                todslices = []
                itemplates = []
                last_obs = iobs
                last_det = det
                last_ref = signal[iobs, det, :]
            todslices.append(todslice)
            itemplates.append(itemplate)
        if len(todslices) != 0:
            project_signal_offsets(
                last_ref, todslices, offset_amplitudes, np.array(itemplates)
            )
        return

    @function_timer
    def add_prior(self, amplitudes_in, amplitudes_out):
        if not self.use_noise_prior:
            return
        offset_amplitudes_in = amplitudes_in[self.name]
        offset_amplitudes_out = amplitudes_out[self.name]
        for iobs, obs in enumerate(self.data.obs):
            tod = obs["tod"]
            for det in tod.local_dets:
                slices = self.offset_slices[iobs][det]
                filters = self.filters[iobs][det]
                for (offsetslice, sigmasqs), noisefilter in zip(slices, filters):
                    amps_in = offset_amplitudes_in[offsetslice]
                    # scipy.signal.convolve will use either `convolve` or `fftconvolve`
                    # depending on the size of the inputs
                    amps_out = scipy.signal.convolve(amps_in, noisefilter, mode="same")
                    offset_amplitudes_out[offsetslice] += amps_out
        return

    @function_timer
    def apply_precond(self, amplitudes_in, amplitudes_out):
        offset_amplitudes_in = amplitudes_in[self.name]
        offset_amplitudes_out = amplitudes_out[self.name]
        if self.use_noise_prior:
            # C_a preconditioner
            for iobs, obs in enumerate(self.data.obs):
                tod = obs["tod"]
                for det in tod.local_dets:
                    slices = self.offset_slices[iobs][det]
                    preconditioners = self.preconditioners[iobs][det]
                    for (offsetslice, sigmasqs), preconditioner in zip(
                        slices, preconditioners
                    ):
                        amps_in = offset_amplitudes_in[offsetslice]
                        if self.precond_width <= 1:
                            # Use C_a prior
                            # scipy.signal.convolve will use either `convolve` or `fftconvolve`
                            # depending on the size of the inputs
                            amps_out = scipy.signal.convolve(
                                amps_in, preconditioner, mode="same"
                            )
                        else:
                            # Use pre-computed Cholesky decomposition
                            amps_out = scipy.linalg.cho_solve_banded(
                                preconditioner,
                                amps_in,
                                overwrite_b=False,
                                check_finite=True,
                            )
                        offset_amplitudes_out[offsetslice] = amps_out
        else:
            # Diagonal preconditioner
            offset_amplitudes_out[:] = offset_amplitudes_in
            for itemplate, iobs, det, todslice, sigmasq in self.offset_templates:
                offset_amplitudes_out[itemplate] *= sigmasq
        return


class TemplateMatrix(TOASTMatrix):
    def __init__(self, data, comm, templates=None):
        """Initialize the template matrix with a given baseline length"""
        self.data = data
        self.comm = comm
        self.templates = []
        for template in templates:
            self.register_template(template)
        return

    @function_timer
    def register_template(self, template):
        """Add template to the list of templates to fit"""
        self.templates.append(template)

    @function_timer
    def apply(self, amplitudes):
        """Compute and return y = F.a"""
        new_signal = self.zero_signal()
        for template in self.templates:
            template.add_to_signal(new_signal, amplitudes)
        return new_signal

    @function_timer
    def apply_transpose(self, signal):
        """Compute and return a = F^T.y"""
        new_amplitudes = self.zero_amplitudes()
        for template in self.templates:
            template.project_signal(signal, new_amplitudes)
        return new_amplitudes

    @function_timer
    def add_prior(self, amplitudes, new_amplitudes):
        """Compute a' += C_a^{-1}.a"""
        for template in self.templates:
            template.add_prior(amplitudes, new_amplitudes)
        return

    @function_timer
    def apply_precond(self, amplitudes):
        """Compute a' = M^{-1}.a"""
        new_amplitudes = self.zero_amplitudes()
        for template in self.templates:
            template.apply_precond(amplitudes, new_amplitudes)
        return new_amplitudes

    @function_timer
    def zero_amplitudes(self):
        """Return a null amplitudes object"""
        new_amplitudes = TemplateAmplitudes(self.templates, self.comm)
        return new_amplitudes

    @function_timer
    def zero_signal(self):
        """Return a distributed vector of signal set to zero.

        The zero signal object will use the same TOD objects but different cache prefix
        """
        new_signal = Signal(self.data, temporary=True, init_val=0)
        return new_signal

    @function_timer
    def clean_signal(self, signal, amplitudes, in_place=True):
        """Clean the given distributed signal vector by subtracting
        the templates multiplied by the given amplitudes.
        """
        # DEBUG begin
        """
        import pdb
        import matplotlib.pyplot as plt
        plt.figure(figsize=[18, 12])
        for sig in [signal]:
            tod = sig.data.obs[0]["tod"]
            for idet, det in enumerate(tod.local_dets):
                plt.subplot(2, 2, idet + 1)
                plt.plot(tod.local_signal(det, sig.name), label=sig.name, zorder=50)
        """
        # DEBUG end
        if in_place:
            outsignal = signal
        else:
            outsignal = signal.copy()
        template_tod = self.apply(amplitudes)
        outsignal -= template_tod
        # DEBUG begin
        """
        for sig, zorder in [(template_tod, 100), (outsignal, 0)]:
            tod = sig.data.obs[0]["tod"]
            for idet, det in enumerate(tod.local_dets):
                plt.subplot(2, 2, idet + 1)
                plt.plot(tod.local_signal(det, sig.name), label=sig.name, zorder=zorder)
        plt.legend(loc="best")
        plt.savefig("test.png")
        plt.close()
        #pdb.set_trace()
        """
        # DEBUG end
        return outsignal


class TemplateAmplitudes(TOASTVector):
    """TemplateAmplitudes objects hold local and shared template amplitudes"""

    def __init__(self, templates, comm):
        self.comm = comm
        self.amplitudes = OrderedDict()
        self.comms = OrderedDict()
        for template in templates:
            self.amplitudes[template.name] = np.zeros(template.namplitude)
            self.comms[template.name] = template.comm
        return

    @function_timer
    def __str__(self):
        result = "template amplitudes:"
        for name, values in self.amplitudes.items():
            result += '\n"{}" : \n{}'.format(name, values)
        return result

    @function_timer
    def dot(self, other):
        """Compute the dot product between the two amplitude vectors"""
        total = 0
        for name, values in self.amplitudes.items():
            comm = self.comms[name]
            if comm is None or comm.rank == 0:
                total += np.dot(values, other.amplitudes[name])
        if self.comm is not None:
            total = self.comm.allreduce(total, op=MPI.SUM)
        return total

    @function_timer
    def __getitem__(self, key):
        return self.amplitudes[key]

    @function_timer
    def __setitem__(self, key, value):
        self.amplitudes[name][:] = value
        return

    @function_timer
    def copy(self):
        new_amplitudes = TemplateAmplitudes([], self.comm)
        for name, values in self.amplitudes.items():
            new_amplitudes.amplitudes[name] = self.amplitudes[name].copy()
            new_amplitudes.comms[name] = self.comms[name]
        return new_amplitudes

    @function_timer
    def __iadd__(self, other):
        """Add the provided amplitudes to this one"""
        if isinstance(other, TemplateAmplitudes):
            for name, values in self.amplitudes.items():
                values += other.amplitudes[name]
        else:
            for name, values in self.amplitudes.items():
                values += other
        return self

    @function_timer
    def __isub__(self, other):
        """Subtract the provided amplitudes from this one"""
        if isinstance(other, TemplateAmplitudes):
            for name, values in self.amplitudes.items():
                values -= other.amplitudes[name]
        else:
            for name, values in self.amplitudes.items():
                values -= other
        return self

    @function_timer
    def __imul__(self, other):
        """Scale the amplitudes"""
        for name, values in self.amplitudes.items():
            values *= other
        return self

    @function_timer
    def __itruediv__(self, other):
        """Divide the amplitudes"""
        for name, values in self.amplitudes.items():
            values /= other
        return self


class TemplateCovariance(TOASTMatrix):
    def __init__(self):
        pass


class ProjectionMatrix(TOASTMatrix):
    """Projection matrix:
        Z = I - P (P^T N^{-1} P)^{-1} P^T N^{-1}
          = I - P B,
    where
         `P` is the pointing matrix
         `N` is the noise matrix and
         `B` is the binning operator
    """

    def __init__(
        self,
        data,
        comm,
        detweights,
        nnz,
        white_noise_cov_matrix,
        common_flag_mask=1,
        flag_mask=1,
    ):
        self.data = data
        self.comm = comm
        self.detweights = detweights
        self.dist_map = DistPixels(data, comm=self.comm, nnz=nnz, dtype=np.float64)
        self.white_noise_cov_matrix = white_noise_cov_matrix
        self.common_flag_mask = common_flag_mask
        self.flag_mask = flag_mask

    @function_timer
    def apply(self, signal):
        """Return Z.y"""
        self.bin_map(signal.name)
        new_signal = signal.copy()
        scanned_signal = Signal(self.data, temporary=True, init_val=0)
        self.scan_map(scanned_signal.name)
        new_signal -= scanned_signal
        return new_signal

    @function_timer
    def bin_map(self, name):
        if self.dist_map.data is not None:
            self.dist_map.data.fill(0.0)
        # FIXME: OpAccumDiag should support separate detweights for each observation
        build_dist_map = OpAccumDiag(
            zmap=self.dist_map,
            name=name,
            detweights=self.detweights[0],
            common_flag_mask=self.common_flag_mask,
            flag_mask=self.flag_mask,
        )
        build_dist_map.exec(self.data)
        self.dist_map.allreduce()
        covariance_apply(self.white_noise_cov_matrix, self.dist_map)
        return

    @function_timer
    def scan_map(self, name):
        scansim = OpSimScan(input_map=self.dist_map, out=name)
        scansim.exec(self.data)
        return


class NoiseMatrix(TOASTMatrix):
    def __init__(
        self, comm, detweights, weightmap=None, common_flag_mask=1, flag_mask=1
    ):
        self.comm = comm
        self.detweights = detweights
        self.weightmap = weightmap
        self.common_flag_mask = common_flag_mask
        self.flag_mask = flag_mask

    @function_timer
    def apply(self, signal, in_place=False):
        """Multiplies the signal with N^{-1}.

        Note that the quality flags cause the corresponding diagonal
        elements of N^{-1} to be zero.
        """
        if in_place:
            new_signal = signal
        else:
            new_signal = signal.copy()
        for iobs, detweights in enumerate(self.detweights):
            for det, detweight in detweights.items():
                new_signal[iobs, det, :] *= detweight
        # Set flagged samples to zero
        new_signal.apply_flags(self.common_flag_mask, self.flag_mask)
        # Scale the signal with the weight map
        new_signal.apply_weightmap(self.weightmap)
        return new_signal

    def apply_transpose(self, signal):
        # Symmetric matrix
        return self.apply(signal)


class PointingMatrix(TOASTMatrix):
    def __init__(self):
        pass


class Signal(TOASTVector):
    """Signal class wraps the TOAST data object but represents only
    one cached signal flavor.
    """

    def __init__(self, data, name=None, init_val=None, temporary=False):
        self.data = data
        self.temporary = temporary
        if self.temporary:
            self.name = get_temporary_name()
        else:
            self.name = name
        if init_val is not None:
            cacheinit = OpCacheInit(name=self.name, init_val=init_val)
            cacheinit.exec(data)
        return

    def __del__(self):
        if self.temporary:
            cacheclear = OpCacheClear(self.name)
            cacheclear.exec(self.data)
            free_temporary_name(self.name)
        return

    @function_timer
    def apply_flags(self, common_flag_mask, flag_mask):
        """Set the signal at flagged samples to zero"""
        flags_apply = OpFlagsApply(
            name=self.name, common_flag_mask=common_flag_mask, flag_mask=flag_mask
        )
        flags_apply.exec(self.data)
        return

    @function_timer
    def apply_weightmap(self, weightmap):
        """Scale the signal with the provided weight map"""
        if weightmap is None:
            return
        scanscale = OpScanScale(distmap=weightmap, name=self.name)
        scanscale.exec(self.data)
        return

    @function_timer
    def copy(self):
        """Return a new Signal object with independent copies of the
        signal vectors.
        """
        new_signal = Signal(self.data, temporary=True)
        copysignal = OpCacheCopy(self.name, new_signal.name, force=True)
        copysignal.exec(self.data)
        return new_signal

    @function_timer
    def __getitem__(self, key):
        """Return a reference to a slice of TOD cache"""
        iobs, det, todslice = key
        tod = self.data.obs[iobs]["tod"]
        return tod.local_signal(det, self.name)[todslice]

    @function_timer
    def __setitem__(self, key, value):
        """Set slice of TOD cache"""
        iobs, det, todslice = key
        tod = self.data.obs[iobs]["tod"]
        tod.local_signal(det, self.name)[todslice] = value
        return

    @function_timer
    def __iadd__(self, other):
        """Add the provided Signal object to this one"""
        for iobs, obs in enumerate(self.data.obs):
            tod = obs["tod"]
            for det in tod.local_dets:
                if isinstance(other, Signal):
                    self[iobs, det, :] += other[iobs, det, :]
                else:
                    self[iobs, det, :] += other
        return self

    @function_timer
    def __isub__(self, other):
        """Subtract the provided Signal object from this one"""
        for iobs, obs in enumerate(self.data.obs):
            tod = obs["tod"]
            for det in tod.local_dets:
                if isinstance(other, Signal):
                    self[iobs, det, :] -= other[iobs, det, :]
                else:
                    self[iobs, det, :] -= other
        return self

    @function_timer
    def __imul__(self, other):
        """Scale the signal"""
        for iobs, obs in enumerate(self.data.obs):
            tod = obs["tod"]
            for det in tod.local_dets:
                self[iobs, det, :] *= other
        return self

    @function_timer
    def __itruediv__(self, other):
        """Divide the signal"""
        for iobs, obs in enumerate(self.data.obs):
            tod = obs["tod"]
            for det in tod.local_dets:
                self[iobs, det, :] /= other
        return self


class PCGSolver:
    """Solves `x` in A.x = b"""

    def __init__(
        self,
        comm,
        templates,
        noise,
        projection,
        signal,
        niter_min=3,
        niter_max=100,
        convergence_limit=1e-12,
    ):
        self.comm = comm
        if comm is None:
            self.rank = 0
        else:
            self.rank = comm.rank
        self.templates = templates
        self.noise = noise
        self.projection = projection
        self.signal = signal
        self.niter_min = niter_min
        self.niter_max = niter_max
        self.convergence_limit = convergence_limit

        self.rhs = self.templates.apply_transpose(
            self.noise.apply(self.projection.apply(self.signal))
        )
        # print("RHS {}: {}".format(self.signal.name, self.rhs))  # DEBUG
        return

    @function_timer
    def apply_lhs(self, amplitudes):
        """Return A.x"""
        new_amplitudes = self.templates.apply_transpose(
            self.noise.apply(self.projection.apply(self.templates.apply(amplitudes)))
        )
        self.templates.add_prior(amplitudes, new_amplitudes)
        return new_amplitudes

    @function_timer
    def solve(self):
        """Standard issue PCG solution of A.x = b

        Returns:
            x : the least squares solution
        """
        log = Logger.get()
        timer0 = Timer()
        timer0.start()
        timer = Timer()
        timer.start()
        # Initial guess is zero amplitudes
        guess = self.templates.zero_amplitudes()
        # print("guess:", guess)  # DEBUG
        # print("RHS:", self.rhs)  # DEBUG
        residual = self.rhs.copy()
        # print("residual(1):", residual)  # DEBUG
        residual -= self.apply_lhs(guess)
        # print("residual(2):", residual)  # DEBUG
        precond_residual = self.templates.apply_precond(residual)
        proposal = precond_residual.copy()
        sqsum = precond_residual.dot(residual)
        init_sqsum, best_sqsum, last_best = sqsum, sqsum, sqsum
        if self.rank == 0:
            log.info("Initial residual: {}".format(init_sqsum))
        # Iterate to convergence
        for iiter in range(self.niter_max):
            if not np.isfinite(sqsum):
                raise RuntimeError("Residual is not finite")
            alpha = sqsum
            alpha /= proposal.dot(self.apply_lhs(proposal))
            alpha_proposal = proposal.copy()
            alpha_proposal *= alpha
            guess += alpha_proposal
            residual -= self.apply_lhs(alpha_proposal)
            del alpha_proposal
            # Prepare for next iteration
            precond_residual = self.templates.apply_precond(residual)
            beta = 1 / sqsum
            # Check for convergence
            sqsum = precond_residual.dot(residual)
            if self.rank == 0:
                timer.report_clear(
                    "Iter = {:4} relative residual: {:12.4e}".format(
                        iiter, sqsum / init_sqsum
                    )
                )
            if sqsum < init_sqsum * self.convergence_limit or sqsum < 1e-30:
                if self.rank == 0:
                    timer0.report_clear(
                        "PCG converged after {} iterations".format(iiter)
                    )
                break
            best_sqsum = min(sqsum, best_sqsum)
            if iiter % 10 == 0 and iiter >= self.niter_min:
                if last_best < best_sqsum * 2:
                    if self.rank == 0:
                        timer0.report_clear(
                            "PCG stalled after {} iterations".format(iiter)
                        )
                    break
                last_best = best_sqsum
            # Select the next direction
            beta *= sqsum
            proposal *= beta
            proposal += precond_residual
        # log.info("{} : Solution: {}".format(self.rank, guess))  # DEBUG
        return guess


class OpMapMaker(Operator):

    # Choose one bit in the common flags for storing gap information
    gap_bit = 2 ** 7
    # Choose one bit in the quality flags for storing processing mask
    mask_bit = 2 ** 7

    def __init__(
        self,
        nside=64,
        nnz=3,
        name=None,
        outdir="out",
        outprefix="",
        write_hits=True,
        zip_maps=False,
        write_wcov_inv=True,
        write_wcov=True,
        write_binned=True,
        write_destriped=True,
        write_rcond=True,
        rcond_limit=1e-3,
        baseline_length=100000,
        maskfile=None,
        weightmapfile=None,
        common_flag_mask=1,
        flag_mask=1,
        intervals="intervals",
        subharmonic_order=None,
        fourier2D_order=None,
        fourier2D_subharmonics=False,
        iter_min=3,
        iter_max=100,
        use_noise_prior=True,
        precond_width=20,
        pixels="pixels",
    ):
        self.nside = nside
        self.npix = 12 * self.nside ** 2
        self.name = name
        self.nnz = nnz
        self.ncov = self.nnz * (self.nnz + 1) // 2
        self.outdir = outdir
        self.outprefix = outprefix
        self.write_hits = write_hits
        self.zip_maps = zip_maps
        self.write_wcov_inv = write_wcov_inv
        self.write_wcov = write_wcov
        self.write_binned = write_binned
        self.write_destriped = write_destriped
        self.write_rcond = write_rcond
        self.rcond_limit = rcond_limit
        self.baseline_length = baseline_length
        self.maskfile = maskfile
        self.weightmap = None
        self.weightmapfile = weightmapfile
        self.common_flag_mask = common_flag_mask
        self.flag_mask = flag_mask
        self.intervals = intervals
        self.subharmonic_order = subharmonic_order
        self.fourier2D_order = fourier2D_order
        self.fourier2D_subharmonics = fourier2D_subharmonics
        self.iter_min = iter_min
        self.iter_max = iter_max
        self.use_noise_prior = use_noise_prior
        self.precond_width = precond_width
        self.pixels = pixels

    def report_timing(self):
        # gt.stop_all()
        all_timers = gather_timers(comm=self.comm)
        names = OrderedDict()
        names["OpMapMaker.exec"] = OrderedDict(
            [
                ("OpMapMaker.flag_gaps", None),
                ("OpMapMaker.get_detweights", None),
                ("OpMapMaker.initialize_binning", None),
                ("OpMapMaker.bin_map", None),
                ("OpMapMaker.load_mask", None),
                ("OpMapMaker.load_weightmap", None),
                ("OpMapMaker.get_templatematrix", None),
                ("OpMapMaker.get_noisematrix", None),
                ("OpMapMaker.get_projectionmatrix", None),
                ("OpMapMaker.get_solver", None),
                (
                    "PCGSolver.solve",
                    OrderedDict(
                        [
                            ("TemplateMatrix.zero_amplitudes", None),
                            ("PCGSolver.apply_lhs", None),
                            ("TemplateMatrix.apply_precond", None),
                        ]
                    ),
                ),
                ("TemplateMatrix.clean_signal", None),
            ]
        )
        names["OpMapMaker.exec"]["PCGSolver.solve"][
            "PCGSolver.apply_lhs"
        ] = OrderedDict(
            [
                (
                    "TemplateMatrix.apply_transpose",
                    OrderedDict(
                        [
                            ("OffsetTemplate.project_signal", None),
                            ("SubharmonicTemplate.project_signal", None),
                            ("fourier2DTemplate.project_signal", None),
                        ]
                    ),
                ),
                ("NoiseMatrix.apply", None),
                (
                    "ProjectionMatrix.apply",
                    OrderedDict(
                        [
                            (
                                "ProjectionMatrix.bin_map",
                                OrderedDict(
                                    [
                                        (
                                            "OpAccumDiag.exec",
                                            OrderedDict(
                                                [
                                                    (
                                                        "OpAccumDiag.exec.apply_flags",
                                                        None,
                                                    ),
                                                    (
                                                        "OpAccumDiag.exec.global_to_local",
                                                        None,
                                                    ),
                                                    ("cov_accum_zmap", None),
                                                ]
                                            ),
                                        ),
                                        ("covariance_apply", None),
                                    ]
                                ),
                            ),
                            (
                                "ProjectionMatrix.scan_map",
                                OrderedDict(
                                    [
                                        (
                                            "OpSimScan.exec",
                                            OrderedDict(
                                                [
                                                    (
                                                        "OpSimScan.exec.global_to_local",
                                                        None,
                                                    ),
                                                    ("OpSimScan.exec.scan_map", None),
                                                ]
                                            ),
                                        )
                                    ]
                                ),
                            ),
                        ]
                    ),
                ),
                (
                    "TemplateMatrix.apply",
                    OrderedDict(
                        [
                            ("OffsetTemplate.add_to_signal", None),
                            ("SubharmonicTemplate.add_to_signal", None),
                            ("fourier2DTemplate.add_to_signal", None),
                        ]
                    ),
                ),
                ("TemplateMatrix.add_prior", None),
            ]
        )
        if self.rank == 0:
            print("all_timers:", all_timers)  # DEBUG

            def report_line(name, indent):
                full_name = name
                if full_name not in all_timers:
                    full_name += " (function_timer)"
                if full_name not in all_timers:
                    return
                t = all_timers[full_name]["time_max"]
                print(indent, "{:.<60}{:8.1f}".format(name, t))
                return

            def report(names, indent):
                if names is None:
                    return
                if isinstance(names, str):
                    report_line(names, indent)
                else:
                    for name, entries in names.items():
                        report_line(name, indent)
                        report(entries, " " * 8 + indent)

            report(names, "-")
            print(flush=True)
        return

    @function_timer
    def get_noisematrix(self, data):
        timer = Timer()
        timer.start()
        noise = NoiseMatrix(
            self.comm,
            self.detweights,
            self.weightmap,
            common_flag_mask=(self.common_flag_mask | self.gap_bit),
            flag_mask=(self.flag_mask | self.mask_bit),
        )
        if self.rank == 0:
            timer.report_clear("Initialize projection matrix")
        return noise

    @function_timer
    def get_projectionmatrix(self, data):
        timer = Timer()
        timer.start()
        projection = ProjectionMatrix(
            data,
            self.comm,
            self.detweights,
            self.nnz,
            self.white_noise_cov_matrix,
            common_flag_mask=(self.common_flag_mask | self.gap_bit),
            # Do not add mask_bit here since it is not
            # included in the white noise matrices
            flag_mask=self.flag_mask,
        )
        if self.rank == 0:
            timer.report_clear("Initialize projection matrix")
        return projection

    @function_timer
    def get_templatematrix(self, data):
        timer = Timer()
        timer.start()
        log = Logger.get()
        templatelist = []
        if self.baseline_length is not None:
            if self.rank == 0:
                log.info(
                    "Initializing offset template, step_length = {}".format(
                        self.baseline_length
                    )
                )
            templatelist.append(
                OffsetTemplate(
                    data,
                    self.detweights,
                    step_length=self.baseline_length,
                    intervals=self.intervals,
                    common_flag_mask=(self.common_flag_mask | self.gap_bit),
                    flag_mask=(self.flag_mask | self.mask_bit),
                    use_noise_prior=self.use_noise_prior,
                    precond_width=self.precond_width,
                )
            )
        if self.subharmonic_order is not None:
            if self.rank == 0:
                log.info(
                    "Initializing subharmonic template, order = {}".format(
                        self.subharmonic_order
                    )
                )
            templatelist.append(
                SubharmonicTemplate(
                    data,
                    self.detweights,
                    order=self.subharmonic_order,
                    intervals=self.intervals,
                    common_flag_mask=(self.common_flag_mask | self.gap_bit),
                    flag_mask=(self.flag_mask | self.mask_bit),
                )
            )
        if self.fourier2D_order is not None:
            log.info(
                "Initializing fourier2D template, order = {}, subharmonics = {}".format(
                    self.fourier2D_order,
                    self.fourier2D_subharmonics,
                )
            )
            templatelist.append(
                Fourier2DTemplate(
                    data,
                    self.detweights,
                    order=self.fourier2D_order,
                    fit_subharmonics=self.fourier2D_subharmonics,
                    intervals=self.intervals,
                    common_flag_mask=(self.common_flag_mask | self.gap_bit),
                    flag_mask=(self.flag_mask | self.mask_bit),
                )
            )
        if len(templatelist) == 0:
            if self.rank == 0:
                log.info("No templates to fit, no destriping done.")
            templates = None
        else:
            templates = TemplateMatrix(data, self.comm, templatelist)
        if self.rank == 0:
            timer.report_clear("Initialize templates")
        return templates

    @function_timer
    def get_solver(self, data, templates, noise, projection, signal):
        timer = Timer()
        timer.start()
        solver = PCGSolver(
            self.comm,
            templates,
            noise,
            projection,
            signal,
            niter_min=self.iter_min,
            niter_max=self.iter_max,
        )
        if self.rank == 0:
            timer.report_clear("Initialize PCG solver")
        return solver

    @function_timer
    def load_mask(self, data):
        """Load processing mask and generate appropriate flag bits"""
        if self.maskfile is None:
            return
        log = Logger.get()
        timer = Timer()
        timer.start()
        if self.rank == 0 and not os.path.isfile(self.maskfile):
            raise RuntimeError(
                "Processing mask does not exist: {}".format(self.maskfile)
            )
        distmap = DistPixels(data, comm=self.comm, nnz=1, dtype=np.float32)
        distmap.read_healpix_fits(self.maskfile)
        if self.rank == 0:
            timer.report_clear("Read processing mask from {}".format(self.maskfile))

        scanmask = OpScanMask(distmap=distmap, flagmask=self.mask_bit)
        scanmask.exec(data)

        if self.rank == 0:
            timer.report_clear("Apply processing mask")

        return

    @function_timer
    def load_weightmap(self, data):
        """Load weight map"""
        if self.weightmapfile is None:
            return
        log = Logger.get()
        timer = Timer()
        timer.start()
        if self.rank == 0 and not os.path.isfile(self.weightmapfile):
            raise RuntimeError(
                "Weight map does not exist: {}".format(self.weightmapfile)
            )
        self.weightmap = DistPixels(data, comm=self.comm, nnz=1, dtype=np.float32)
        self.weightmap.read_healpix_fits(self.weightmapfile)
        if self.rank == 0:
            timer.report_clear("Read weight map from {}".format(self.weightmapfile))
        return

    @function_timer
    def exec(self, data, comm=None):
        log = Logger.get()
        timer = Timer()

        # Initialize objects
        if comm is None:
            self.comm = data.comm.comm_world
        else:
            self.comm = comm
        if self.comm is None:
            self.rank = 0
        else:
            self.rank = self.comm.rank
        self.flag_gaps(data)
        self.get_detweights(data)
        self.initialize_binning(data)
        if self.write_binned:
            self.bin_map(data, "binned")
        self.load_mask(data)
        self.load_weightmap(data)

        # Solve template amplitudes

        templates = self.get_templatematrix(data)
        if templates is None:
            return
        noise = self.get_noisematrix(data)
        projection = self.get_projectionmatrix(data)
        signal = Signal(data, name=self.name)
        solver = self.get_solver(data, templates, noise, projection, signal)
        timer.start()
        amplitudes = solver.solve()
        if self.rank == 0:
            timer.report_clear("Solve amplitudes")

        # Clean TOD
        templates.clean_signal(signal, amplitudes)
        if self.rank == 0:
            timer.report_clear("Clean TOD")

        if self.write_destriped:
            self.bin_map(data, "destriped")

        return

    @function_timer
    def flag_gaps(self, data):
        """Add flag bits between the intervals"""
        timer = Timer()
        timer.start()
        flag_gaps = OpFlagGaps(common_flag_value=self.gap_bit, intervals=self.intervals)
        flag_gaps.exec(data)
        if self.rank == 0:
            timer.report_clear("Flag gaps")
        return

    @function_timer
    def bin_map(self, data, suffix):
        log = Logger.get()
        timer = Timer()

        dist_map = DistPixels(data, comm=self.comm, nnz=self.nnz, dtype=np.float64)
        if dist_map.data is not None:
            dist_map.data.fill(0.0)
        # FIXME: OpAccumDiag should support separate detweights for each observation
        build_dist_map = OpAccumDiag(
            zmap=dist_map,
            name=self.name,
            detweights=self.detweights[0],
            common_flag_mask=(self.common_flag_mask | self.gap_bit),
            flag_mask=self.flag_mask,
        )
        build_dist_map.exec(data)
        dist_map.allreduce()
        if self.rank == 0:
            timer.report_clear("  Build noise-weighted map")

        covariance_apply(self.white_noise_cov_matrix, dist_map)
        if self.rank == 0:
            timer.report_clear("  Apply noise covariance")

        fname = os.path.join(self.outdir, self.outprefix + suffix + ".fits")
        if self.zip_maps:
            fname += ".gz"
        dist_map.write_healpix_fits(fname)
        if self.rank == 0:
            timer.report_clear("  Write map to {}".format(fname))

        return

    @function_timer
    def get_detweights(self, data):
        """Each observation will have its own detweight dictionary"""
        timer = Timer()
        timer.start()
        self.detweights = []
        for obs in data.obs:
            tod = obs["tod"]
            if "noise" in obs:
                noise = obs["noise"]
            else:
                noise = None
            detweights = {}
            for det in tod.local_dets:
                if noise is None:
                    noisevar = 1
                else:
                    # Determine an approximate white noise level,
                    # accounting for the fact that the PSD may have a
                    # transfer function roll-off near Nyquist
                    freq = noise.freq(det)
                    psd = noise.psd(det)
                    rate = noise.rate(det)
                    ind = np.logical_and(freq > rate * 0.2, freq < rate * 0.4)
                    noisevar = np.median(psd[ind])
                detweights[det] = 1 / noisevar
            self.detweights.append(detweights)
        if self.rank == 0:
            timer.report_clear("Get detector weights")
        return

    @function_timer
    def initialize_binning(self, data):
        log = Logger.get()
        timer = Timer()
        timer.start()

        if self.rank == 0:
            os.makedirs(self.outdir, exist_ok=True)

        self.white_noise_cov_matrix = DistPixels(
            data, comm=self.comm, nnz=self.ncov, dtype=np.float64
        )
        if self.white_noise_cov_matrix.data is not None:
            self.white_noise_cov_matrix.data.fill(0.0)

        hits = DistPixels(data, comm=self.comm, nnz=1, dtype=np.int64)
        if hits.data is not None:
            hits.data.fill(0)

        # compute the hits and covariance once, since the pointing and noise
        # weights are fixed.
        # FIXME: OpAccumDiag should support separate weights for each observation

        build_wcov = OpAccumDiag(
            detweights=self.detweights[0],
            invnpp=self.white_noise_cov_matrix,
            hits=hits,
            common_flag_mask=(self.common_flag_mask | self.gap_bit),
            flag_mask=self.flag_mask,
        )
        build_wcov.exec(data)

        if self.comm is not None:
            self.comm.Barrier()
        if self.rank == 0:
            timer.report_clear("Accumulate N_pp'^1")

        self.white_noise_cov_matrix.allreduce()

        if self.comm is not None:
            self.comm.Barrier()
        if self.rank == 0:
            timer.report_clear("All reduce N_pp'^1")

        if self.write_hits:
            hits.allreduce()
            fname = os.path.join(self.outdir, self.outprefix + "hits.fits")
            if self.zip_maps:
                fname += ".gz"
            hits.write_healpix_fits(fname)
            if self.rank == 0:
                log.info("Wrote hits to {}".format(fname))
            if self.rank == 0:
                timer.report_clear("Write hits")

        if self.write_wcov_inv:
            fname = os.path.join(self.outdir, self.outprefix + "invnpp.fits")
            if self.zip_maps:
                fname += ".gz"
            self.white_noise_cov_matrix.write_healpix_fits(fname)
            if self.rank == 0:
                log.info("Wrote inverse white noise covariance to {}".format(fname))
            if self.rank == 0:
                timer.report_clear("Write N_pp'^1")

        if self.write_rcond:
            # Reciprocal condition numbers
            rcond = covariance_rcond(self.white_noise_cov_matrix)
            if self.rank == 0:
                timer.report_clear("Compute reciprocal condition numbers")
            fname = os.path.join(self.outdir, self.outprefix + "rcond.fits")
            if self.zip_maps:
                fname += ".gz"
            rcond.write_healpix_fits(fname)
            if self.rank == 0:
                log.info("Wrote reciprocal condition numbers to {}".format(fname))
            if self.rank == 0:
                timer.report_clear("Write rcond")

        # Invert the white noise covariance in each pixel
        covariance_invert(self.white_noise_cov_matrix, self.rcond_limit)
        if self.rank == 0:
            timer.report_clear("Invert N_pp'^1")

        if self.write_wcov:
            fname = os.path.join(self.outdir, self.outprefix + "npp.fits")
            if self.zip_maps:
                fname += ".gz"
            self.white_noise_cov_matrix.write_healpix_fits(fname)
            if self.rank == 0:
                log.info("Wrote white noise covariance to {}".format(fname))
            if self.rank == 0:
                timer.report_clear("Write N_pp'")

        return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

# import functions into our public API

from .tod import TOD, TODCache

from .interval import Interval, OpFlagGaps

from .tod_math import (
    calibrate,
    sim_noise_timestream,
    OpCacheCopy,
    OpCacheClear,
    flagged_running_average,
    OpCacheInit,
    OpFlagsApply,
)

from .sim_noise import AnalyticNoise

from .sim_interval import regular_intervals

from .sim_det_noise import OpSimNoise

from .sim_focalplane import (
    hex_layout,
    rhombus_layout,
    hex_pol_angles_qu,
    hex_pol_angles_radial,
    rhomb_pol_angles_qu,
    plot_focalplane,
)

from .noise import Noise

from .polyfilter import OpPolyFilter, OpPolyFilter2D

from .gainscrambler import OpGainScrambler
from .applygain import OpApplyGain, write_calibration_file

from .memorycounter import OpMemoryCounter

from .tidas import available as tidas_available

from .spt3g_utils import available as spt3g_available



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

from .psd_math import autocov_psd, crosscov_psd

from .noise_estimation import OpNoiseEstim



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

from .utils import (
    Logger,
    AlignedF64,
    AlignedI64,
    ensure_buffer_f64,
    ensure_buffer_i64,
    object_ndim,
)

from ._libtoast import (
    healpix_ang2vec,
    healpix_vec2ang,
    healpix_vecs2angpa,
    HealpixPixels,
)


import numpy as np

from .timing import function_timer


def ang2vec(theta, phi):
    """Convert spherical coordinates to a unit vector.

    The theta angle is measured down from the North pole and phi is
    measured from the prime meridian.

    Args:
        theta (array_like): The spherical coordinate theta angles in
            radians.
        phi (array like): The spherical coordinate phi angles in radians.

    Returns:
        (array):  The array of output vectors.

    """
    intheta = ensure_buffer_f64(theta)
    inphi = ensure_buffer_f64(phi)
    ntheta = len(intheta)
    nphi = len(inphi)
    if ntheta != nphi:
        raise RuntimeError("theta / phi vectors must have the same length")
    vec = AlignedF64(3 * nphi)
    healpix_ang2vec(intheta, inphi, vec)
    if len(vec) == 3:
        if (object_ndim(theta) == 1) or (object_ndim(phi) == 1):
            return vec.array().reshape(1, 3)
        else:
            return vec.array()
    else:
        return vec.array().reshape((-1, 3))


def vec2ang(vec):
    """Convert unit vectors to spherical coordinates.

    The theta angle is measured down from the North pole and phi is
    measured from the prime meridian.

    Args:
        vec (array_like):  The array of input vectors.

    Returns:
        (tuple): The (theta, phi) arrays in radians.

    """
    invec = ensure_buffer_f64(vec)
    n = len(invec) // 3
    theta = AlignedF64(n)
    phi = AlignedF64(n)
    healpix_vec2ang(invec, theta, phi)
    if len(vec) == 3:
        if object_ndim(vec) == 2:
            return (theta.array(), phi.array())
        else:
            return (theta[0], phi[0])
    else:
        return (theta.array(), phi.array())


def vecs2angpa(vec):
    """Convert direction / orientation unit vectors.

    The inputs are flat-packed pairs of direction and orientation unit
    vectors (6 float64 values total per sample).  The outputs are the
    theta, phi, and position angle of the location on the sphere.

    The theta angle is measured down from the North pole and phi is
    measured from the prime meridian.  The position angle is with respect
    to the local meridian at the point described by the theta / phi
    coordinates.

    Args:
        vec (array_like):  The array of packed input direction and
            orientation vectors.

    Returns:
        (tuple): The (theta, phi, pa) arrays in radians.

    """
    invec = ensure_buffer_f64(vec)
    n = len(invec) // 6
    theta = AlignedF64(n)
    phi = AlignedF64(n)
    pa = AlignedF64(n)
    healpix_vecs2angpa(invec, theta, phi, pa)
    if len(vec) == 6:
        if object_ndim(vec) == 2:
            return (theta.array(), phi.array(), pa.array())
        else:
            return (theta[0], phi[0], pa[0])
    else:
        return (theta.array(), phi.array(), pa.array())


class Pixels(object):
    """Class for HEALPix pixel operations at a fixed NSIDE.

    This provides a very thin wrapper around the internal HealpixPixels class.
    This wrapper mainly exists to provide methods that return a newly created
    output buffer (rather than requiring the user to pass in the buffer).  If
    you are happy pre-creating the output buffer then you can call methods
    of HealpixPixels directly.

    Args:
        nside (int): the map NSIDE.

    """

    def __init__(self, nside=1):
        self.hpix = HealpixPixels(nside)

    def __del__(self):
        if self.hpix is not None:
            del self.hpix

    def reset(self, nside):
        """Reset the class instance to use a new NSIDE value.

        Args:
            nside (int): the map NSIDE.

        Returns:
            None

        """
        self.hpix.reset(nside)
        return

    def ang2nest(self, theta, phi):
        """Convert spherical coordinates to pixels in NESTED ordering.

        The theta angle is measured down from the North pole and phi is
        measured from the prime meridian.

        Args:
            theta (array_like): Input spherical coordinate theta angles in
                radians.
            phi (array like): Input spherical coordinate phi angles in
                radians.

        Returns:
            (array): Output pixel indices.

        """
        intheta = ensure_buffer_f64(theta)
        inphi = ensure_buffer_f64(phi)
        ntheta = len(intheta)
        nphi = len(inphi)
        if ntheta != nphi:
            raise RuntimeError("theta / phi vectors must have the same length")
        pix = AlignedI64(nphi)
        self.hpix.ang2nest(intheta, inphi, pix)
        if nphi == 1:
            if (object_ndim(theta) == 1) or (object_ndim(phi) == 1):
                return pix.array()
            else:
                return pix[0]
        else:
            return pix.array()

    def ang2ring(self, theta, phi):
        """Convert spherical coordinates to pixels in RING ordering.

        The theta angle is measured down from the North pole and phi is
        measured from the prime meridian.

        Args:
            theta (array_like): Input spherical coordinate theta angles in
                radians.
            phi (array like): Input spherical coordinate phi angles in
                radians.

        Returns:
            (array): Output pixel indices.

        """
        intheta = ensure_buffer_f64(theta)
        inphi = ensure_buffer_f64(phi)
        ntheta = len(intheta)
        nphi = len(inphi)
        if ntheta != nphi:
            raise RuntimeError("theta / phi vectors must have the same length")
        pix = AlignedI64(nphi)
        self.hpix.ang2ring(intheta, inphi, pix)
        if nphi == 1:
            if (object_ndim(theta) == 1) or (object_ndim(phi) == 1):
                return pix.array()
            else:
                return pix[0]
        else:
            return pix.array()

    def vec2nest(self, vec):
        """Convert unit vectors to pixels in NESTED ordering.

        The theta angle is measured down from the North pole and phi is
        measured from the prime meridian.

        Args:
            vec (array_like): Input packed unit vectors.

        Returns:
            (array): Output pixel indices.

        """
        invec = ensure_buffer_f64(vec)
        n = len(invec) // 3
        pix = AlignedI64(n)
        self.hpix.vec2nest(invec, pix)
        if len(vec) == 3:
            if object_ndim(vec) == 2:
                return pix.array()
            else:
                return pix[0]
        else:
            return pix.array()

    def vec2ring(self, vec):
        """Convert unit vectors to pixels in RING ordering.

        The theta angle is measured down from the North pole and phi is
        measured from the prime meridian.

        Args:
            vec (array_like): Input packed unit vectors.

        Returns:
            (array): Output pixel indices.

        """
        invec = ensure_buffer_f64(vec)
        n = len(invec) // 3
        pix = AlignedI64(n)
        self.hpix.vec2ring(invec, pix)
        if len(vec) == 3:
            if object_ndim(vec) == 2:
                return pix.array()
            else:
                return pix[0]
        else:
            return pix.array()

    def ring2nest(self, ringpix):
        """Convert RING ordered pixel numbers into NESTED ordering.

        Args:
            in (array_like): Input pixel indices.

        Returns:
            (array): Output pixel indices.

        """
        inpix = ensure_buffer_i64(ringpix)
        n = len(inpix)
        out = AlignedI64(n)
        self.hpix.ring2nest(inpix, out)
        if n == 1:
            if object_ndim(ringpix) == 1:
                return out.array()
            else:
                return out[0]
        else:
            return out.array()

    def nest2ring(self, nestpix):
        """Convert NESTED ordered pixel numbers into RING ordering.

        Args:
            in (array_like): Input pixel indices.

        Returns:
            (array): Output pixel indices.

        """
        inpix = ensure_buffer_i64(nestpix)
        n = len(inpix)
        out = AlignedI64(n)
        self.hpix.ring2nest(inpix, out)
        if n == 1:
            if object_ndim(nestpix) == 1:
                return out.array()
            else:
                return out[0]
        else:
            return out.array()

    def degrade_ring(self, factor, inpix):
        """Degrade RING ordered pixel numbers.

        Each 'factor' is one division by two in the NSIDE resolution.  So
        a factor of '3' would divide the NSIDE value by 8.

        Args:
            factor (int):  The degrade factor.
            in (array_like): Input pixel indices.

        Returns:
            (array): Output pixel indices.

        """
        inp = ensure_buffer_i64(inpix)
        n = len(inp)
        out = AlignedI64(n)
        self.hpix.degrade_ring(factor, inp, out)
        if n == 1:
            if object_ndim(inpix) == 1:
                return out.array()
            else:
                return out[0]
        else:
            return out.array()

    def degrade_nest(self, factor, inpix):
        """Degrade NESTED ordered pixel numbers.

        Each 'factor' is one division by two in the NSIDE resolution.  So
        a factor of '3' would divide the NSIDE value by 8.

        Args:
            factor (int):  The degrade factor.
            in (array_like): Input pixel indices.

        Returns:
            (array): Output pixel indices.

        """
        inp = ensure_buffer_i64(inpix)
        n = len(inp)
        out = AlignedI64(n)
        self.hpix.degrade_nest(factor, inp, out)
        if n == 1:
            if object_ndim(inpix) == 1:
                return out.array()
            else:
                return out[0]
        else:
            return out.array()

    def upgrade_ring(self, factor, inpix):
        """Upgrade RING ordered pixel numbers.

        Each 'factor' is one multiplication by two in the NSIDE
        resolution.  So a factor of '3' would multiply the NSIDE value
        by 8.

        Args:
            factor (int):  The upgrade factor.
            inpix (array_like): Input pixel indices.

        Returns:
            (array): Output pixel indices.

        """
        inp = ensure_buffer_i64(inpix)
        n = len(inp)
        out = AlignedI64(n)
        self.hpix.upgrade_ring(factor, inp, out)
        if n == 1:
            if object_ndim(inpix) == 1:
                return out.array()
            else:
                return out[0]
        else:
            return out.array()

    def upgrade_nest(self, factor, inpix):
        """Upgrade RING ordered pixel numbers.

        Each 'factor' is one multiplication by two in the NSIDE
        resolution.  So a factor of '3' would multiply the NSIDE value
        by 8.

        Args:
            factor (int):  The upgrade factor.
            inpix (array_like): Input pixel indices.

        Returns:
            (array): Output pixel indices.

        """
        inp = ensure_buffer_i64(inpix)
        n = len(inp)
        out = AlignedI64(n)
        self.hpix.upgrade_nest(factor, inp, out)
        if n == 1:
            if object_ndim(inpix) == 1:
                return out.array()
            else:
                return out[0]
        else:
            return out.array()



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import numpy as np

from .mpi import Comm


# This is effectively the "Painter's Partition Problem".


def distribute_required_groups(A, max_per_group):
    ngroup = 1
    total = 0
    for i in range(A.shape[0]):
        total += A[i]
        if total > max_per_group:
            total = A[i]
            ngroup += 1
    return ngroup


def distribute_partition(A, k):
    low = np.max(A)
    high = np.sum(A)
    while low < high:
        mid = low + int((high - low) / 2)
        required = distribute_required_groups(A, mid)
        if required <= k:
            high = mid
        else:
            low = mid + 1
    return low


def distribute_discrete(sizes, groups, pow=1.0, breaks=None):
    """Distribute indivisible blocks of items between groups.

    Given some contiguous blocks of items which cannot be
    subdivided, distribute these blocks to the specified
    number of groups in a way which minimizes the maximum
    total items given to any group.  Optionally weight the
    blocks by a power of their size when computing the
    distribution.

    Args:
        sizes (list): The sizes of the indivisible blocks.
        groups (int): The number of groups.
        pow (float): The power to use for weighting
        breaks (list): List of hard breaks in the data distribution.

    Returns:
        A list of tuples.  There is one tuple per group.
        The first element of the tuple is the first item
        assigned to the group, and the second element is
        the number of items assigned to the group.
    """
    chunks = np.array(sizes, dtype=np.int64)
    weights = np.power(chunks.astype(np.float64), pow)
    max_per_proc = float(distribute_partition(weights.astype(np.int64), groups))

    target = np.sum(weights) / groups

    dist = []

    off = 0
    curweight = 0.0

    all_breaks = None
    if breaks is not None:
        # Check that the problem makes sense
        all_breaks = np.unique(breaks)
        all_breaks = all_breaks[all_breaks > 0]
        all_breaks = all_breaks[all_breaks < chunks.size]
        all_breaks = np.sort(all_breaks)
        if all_breaks.size + 1 > groups:
            raise RuntimeError(
                "Cannot divide {} chunks to {} groups with {} breaks.".format(
                    chunks.size, groups, all_breaks.size
                )
            )

    at_break = False
    for cur in range(0, weights.shape[0]):
        if curweight + weights[cur] > max_per_proc or at_break:
            dist.append((off, cur - off))
            over = curweight - target
            curweight = weights[cur] + over
            off = cur
        else:
            curweight += weights[cur]
        if all_breaks is not None:
            at_break = False
            if cur + 1 in all_breaks:
                at_break = True

    dist.append((off, weights.shape[0] - off))

    if len(dist) != groups:
        raise RuntimeError(
            "Number of distributed groups different than " "number requested"
        )
    return dist


def distribute_uniform(totalsize, groups, breaks=None):
    """Uniformly distribute items between groups.

    Given some number of items and some number of groups,
    distribute the items between groups in the most Uniform
    way possible.

    Args:
        totalsize (int): The total number of items.
        groups (int): The number of groups.
        breaks (list): List of hard breaks in the data distribution.

    Returns:
        (list): there is one tuple per group.  The first element of the tuple
            is the first item assigned to the group, and the second element is
            the number of items assigned to the group.

    """
    if breaks is not None:
        all_breaks = np.unique(breaks)
        all_breaks = all_breaks[all_breaks > 0]
        all_breaks = all_breaks[all_breaks < totalsize]
        all_breaks = np.sort(all_breaks)
        if len(all_breaks) > groups - 1:
            raise RuntimeError(
                "Cannot distribute {} chunks with {} breaks over {} groups"
                "".format(totalsize, len(all_breaks), groups)
            )
        groupcounts = []
        groupsizes = []
        offset = 0
        groupsleft = groups
        totalleft = totalsize
        for brk in all_breaks:
            length = brk - offset
            groupcount = int(np.round(groupsleft * length / totalleft))
            groupcount = max(1, groupcount)
            groupcount = min(groupcount, groupsleft)
            groupcounts.append(groupcount)
            groupsizes.append(length)
            groupsleft -= groupcount
            totalleft -= length
            offset = brk
        groupcounts.append(groupsleft)
        groupsizes.append(totalleft)
    else:
        groupcounts = [groups]
        groupsizes = [totalsize]

    dist = []
    offset = 0
    for groupsize, groupcount in zip(groupsizes, groupcounts):
        for i in range(groupcount):
            myn = groupsize // groupcount
            off = 0
            leftover = groupsize % groupcount
            if i < leftover:
                myn = myn + 1
                off = i * myn
            else:
                off = ((myn + 1) * leftover) + (myn * (i - leftover))
            dist.append((offset + off, myn))
        offset += groupsize
    return dist


def distribute_samples(
    mpicomm,
    detectors,
    samples,
    detranks=1,
    detbreaks=None,
    sampsizes=None,
    sampbreaks=None,
):
    """Distribute data by detector and sample.

    Given a list of detectors and some number of samples, distribute
    the data in a load balanced way.  Optionally account for constraints
    on this distribution.  The samples may be grouped by indivisible
    chunks, and there may be forced breaks in the distribution in both
    the detector and chunk directions.

                            samples -->
                      +--------------+--------------
                    / | sampsize[0]  | sampsize[1] ...
        detrank = 0   +--------------+--------------
                    \ | sampsize[0]  | sampsize[1] ...
                      +--------------+--------------
                    / | sampsize[0]  | sampsize[1] ...
        detrank = 1   +--------------+--------------
                    \ | sampsize[0]  | sampsize[1] ...
                      +--------------+--------------
                      | ...

    Args:
        mpicomm (mpi4py.MPI.Comm):  the MPI communicator over which the
            data is distributed.  If None, then all data is assigned to a
            single process.
        detectors (list):  The list of detector names.
        samples (int):  The total number of samples.
        detranks (int):  The dimension of the process grid in the detector
            direction.  The MPI communicator size must be evenly divisible
            by this number.
        detbreaks (list):  Optional list of hard breaks in the detector
            distribution.
        sampsizes (list):  Optional list of sample chunk sizes which
            cannot be split.
        sampbreaks (list):  Optional list of hard breaks in the sample
            distribution.

    Returns:
        tuple of lists: the 3 lists returned contain information about
        the detector distribution, the sample distribution, and the chunk
        distribution.  The first list has one entry for each detrank and
        contains the list of detectors for that row of the process grid.
        The second list contains tuples of (first sample, N samples) for
        each column of the process grid.  The third list contains tuples
        of (first chunk, N chunks) for each column of the process grid.

    """
    nproc = 1
    if mpicomm is not None:
        nproc = mpicomm.size

    if nproc % detranks != 0:
        raise RuntimeError(
            "The number of detranks ({}) does not divide evenly "
            "into the number of processes ({})".format(detranks, nproc)
        )

    # Compute the other dimension of the process grid.
    sampranks = nproc // detranks

    # Distribute detectors uniformly, but respecting forced breaks in the
    # grouping specified by the calling code.

    dist_detsindx = distribute_uniform(len(detectors), detranks, breaks=detbreaks)
    dist_dets = [detectors[d[0] : d[0] + d[1]] for d in dist_detsindx]

    # Distribute samples using both the chunking and the forced breaks

    if sampsizes is not None:
        dist_sizes = distribute_discrete(sampsizes, sampranks, breaks=sampbreaks)
        dist_samples = []
        off = 0
        for ds in dist_sizes:
            cursamp = np.sum(sampsizes[ds[0] : ds[0] + ds[1]])
            dist_samples.append((off, cursamp))
            off += cursamp
    else:
        dist_samples = distribute_uniform(samples, sampranks, breaks=sampbreaks)
        dist_sizes = [(x, 1) for x in range(sampranks)]

    return (dist_dets, dist_samples, dist_sizes)


class Data(object):
    """Class which represents distributed data

    A Data object contains a list of observations assigned to
    each process group in the Comm.

    Args:
        comm (:class:`toast.Comm`): the toast Comm class for distributing the data.

    """

    def __init__(self, comm=Comm()):
        self._comm = comm
        self.obs = []
        """The list of observations.
        """
        self._metadata = {}

    def __contains__(self, key):
        return key in self._metadata

    def __getitem__(self, key):
        return self._metadata[key]

    def __setitem__(self, key, value):
        self._metadata[key] = value

    @property
    def comm(self):
        """The toast.Comm over which the data is distributed."""
        return self._comm

    def clear(self):
        """Clear the list of observations."""
        for ob in self.obs:
            ob.clear()
        return

    def info(self, handle=None, flag_mask=255, common_flag_mask=255, intervals=None):
        """Print information about the distributed data.

        Information is written to the specified file handle.  Only the rank 0
        process writes.  Optional flag masks are used when computing the
        number of good samples.

        Args:
            handle (descriptor):  file descriptor supporting the write()
                method.  If None, use print().
            flag_mask (int):  bit mask to use when computing the number of
                good detector samples.
            common_flag_mask (int):  bit mask to use when computing the
                number of good telescope pointings.
            intervals (str):  optional name of an intervals object to print
                from each observation.

        Returns:
            None

        """
        # Each process group gathers their output

        groupstr = ""
        procstr = ""

        gcomm = self._comm.comm_group
        wcomm = self._comm.comm_world
        rcomm = self._comm.comm_rank

        if wcomm is None:
            msg = "Data distributed over a single process (no MPI)"
            if handle is None:
                print(msg, flush=True)
            else:
                handle.write(msg)
        else:
            if wcomm.rank == 0:
                msg = "Data distributed over {} processes in {} groups\n".format(
                    self._comm.world_size, self._comm.ngroups
                )
                if handle is None:
                    print(msg, flush=True)
                else:
                    handle.write(msg)

        def _get_optional(k, dt):
            if k in dt:
                return dt[k]
            else:
                return None

        for ob in self.obs:
            id = ob["id"]
            tod = _get_optional("tod", ob)
            intrvl = None
            if intervals is not None:
                _get_optional(intervals, ob)

            if self._comm.group_rank == 0:
                groupstr = "observation {}:\n".format(id)
                for ko in sorted(ob.keys()):
                    groupstr = "{}  key {}\n".format(groupstr, ko)
                if tod is not None:
                    groupstr = "{}  {} total samples, {} detectors\n".format(
                        groupstr, tod.total_samples, len(tod.detectors)
                    )
                if intrvl is not None:
                    groupstr = "{}  {} intervals:\n".format(groupstr, len(intrvl))
                    for it in intrvl:
                        groupstr = "{}    {} --> {} ({} --> {})\n".format(
                            groupstr, it.first, it.last, it.start, it.stop
                        )

            # rank zero of the group will print general information,
            # and each process will get its statistics.

            procstr = "  proc {}\n".format(self._comm.group_rank)
            if tod is not None:
                offset, nsamp = tod.local_samples
                dets = tod.local_dets

                my_chunks = 1
                if tod.local_chunks is not None:
                    my_chunks = tod.local_chunks[1]
                procstr = "{}    sample range {} --> {} in {} chunks:\n".format(
                    procstr, offset, (offset + nsamp - 1), my_chunks
                )

                if tod.local_chunks is not None:
                    chkoff = tod.local_samples[0]
                    for chk in range(tod.local_chunks[1]):
                        abschk = tod.local_chunks[0] + chk
                        chkstart = chkoff
                        chkstop = chkstart + tod.total_chunks[abschk] - 1
                        procstr = "{}      {} --> {}\n".format(
                            procstr, chkstart, chkstop
                        )
                        chkoff += tod.total_chunks[abschk]

                if nsamp > 0:
                    stamps = tod.local_times()

                    procstr = "{}    timestamps {} --> {}\n".format(
                        procstr, stamps[0], stamps[-1]
                    )

                    common = tod.local_common_flags()
                    for dt in dets:
                        procstr = "{}    det {}:\n".format(procstr, dt)

                        pdata = tod.local_pointing(dt)

                        procstr = (
                            "{}      pntg [{:.3e} {:.3e} {:.3e} {:.3e}] "
                            "--> [{:.3e} {:.3e} {:.3e} {:.3e}]\n".format(
                                procstr,
                                pdata[0, 0],
                                pdata[0, 1],
                                pdata[0, 2],
                                pdata[0, 3],
                                pdata[-1, 0],
                                pdata[-1, 1],
                                pdata[-1, 2],
                                pdata[-1, 3],
                            )
                        )

                        data = tod.local_signal(dt)
                        flags = tod.local_flags(dt)
                        procstr = "{}      {:.3e} ({}) --> {:.3e} ({})\n".format(
                            procstr, data[0], flags[0], data[-1], flags[-1]
                        )
                        good = np.where(
                            ((flags & flag_mask) | (common & common_flag_mask)) == 0
                        )[0]
                        procstr = "{}        {} good samples\n".format(
                            procstr, len(good)
                        )
                        try:
                            min = np.min(data[good])
                            max = np.max(data[good])
                            mean = np.mean(data[good])
                            rms = np.std(data[good])
                            procstr = (
                                "{}        min = {:.4e}, max = {:.4e},"
                                " mean = {:.4e}, rms = {:.4e}\n".format(
                                    procstr, min, max, mean, rms
                                )
                            )
                        except FloatingPointError:
                            procstr = (
                                "{}        min = N/A, max = N/A, "
                                "mean = N/A, rms = N/A\n".format(procstr)
                            )

                    for cname in tod.cache.keys():
                        procstr = "{}    cache {}:\n".format(procstr, cname)
                        ref = tod.cache.reference(cname)
                        min = np.min(ref)
                        max = np.max(ref)
                        mean = np.mean(ref)
                        rms = np.std(ref)
                        procstr = (
                            "{}        min = {:.4e}, max = {:.4e}, "
                            "mean = {:.4e}, rms = {:.4e}\n".format(
                                procstr, min, max, mean, rms
                            )
                        )

            recvstr = ""
            if self._comm.group_rank == 0:
                groupstr = "{}{}".format(groupstr, procstr)
            if gcomm is not None:
                for p in range(1, self._comm.group_size):
                    if gcomm.rank == 0:
                        recvstr = gcomm.recv(source=p, tag=p)
                        groupstr = "{}{}".format(groupstr, recvstr)
                    elif p == gcomm.rank:
                        gcomm.send(procstr, dest=0, tag=p)
                    gcomm.barrier()

        # the world rank 0 process collects output from all groups and
        # writes to the handle

        recvgrp = ""
        if self._comm.world_rank == 0:
            if handle is None:
                print(groupstr, flush=True)
            else:
                handle.write(groupstr)
        if wcomm is not None:
            for g in range(1, self._comm.ngroups):
                if wcomm.rank == 0:
                    recvgrp = rcomm.recv(source=g, tag=g)
                    if handle is None:
                        print(recvgrp, flush=True)
                    else:
                        handle.write(recvgrp)
                elif g == self._comm.group:
                    if gcomm.rank == 0:
                        rcomm.send(groupstr, dest=0, tag=g)
                wcomm.barrier()
        return

    def split(self, key):
        """Split the Data object.

        Split the Data object based on the value of `key` in the
        observation dictionary.

        Args:
            key(str) :  Observation key to use.

        Returns:
            List of 2-tuples of the form (value, data)

        """
        # Build a superset of all values
        values = set()
        for obs in self.obs:
            if key not in obs:
                raise RuntimeError(
                    'Cannot split data by "{}". Key is not '
                    "defined for all observations.".format(key)
                )
            values.add(obs[key])
        all_values = None
        if self._comm.comm_world is None:
            all_values = [values]
        else:
            all_values = self._comm.comm_world.allgather(values)
        for vals in all_values:
            values = values.union(vals)

        # Order the values alphabetically.
        values = sorted(list(values))

        # Split the data
        datasplit = []
        for value in values:
            new_data = Data(comm=self._comm)
            for obs in self.obs:
                if obs[key] == value:
                    new_data.obs.append(obs)
            datasplit.append((value, new_data))

        return datasplit



# Copyright (c) 2019-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import os
import pickle
import sys

import numpy as np

from ..timing import function_timer, Timer
from ..tod import AnalyticNoise
from ..utils import Logger, Environment
from .. import qarray


XAXIS, YAXIS, ZAXIS = np.eye(3)


def name2id(name, maxval=2 ** 16):
    """Map a name into an index."""
    value = 0
    for c in name:
        value += ord(c)
    return value % maxval


class Focalplane:
    _detweights = None
    _detquats = None
    _noise = None

    def __init__(
        self, detector_data=None, fname_pickle=None, sample_rate=None, radius_deg=None
    ):
        """Instantiate a focalplane

        Args:
            detector_data (dict) :  Dictionary of detector attributes, such
                as detector quaternions and noise parameters.
            fname_pickle (str) :  Pickle file containing the focal
                 plane dictionary.  If both `detector_data` and
                 `fname_pickle` are set, the dictionaries are merged.
            sample_rate (float) :  Default sampling rate for all
                detectors.  Will be overridden by 'fsample' fields
                if they exist for the detectors in the dictionary.
            radius_deg (float) :  force the radius of the focal plane.
                otherwise it will be calculated from the detector
                offsets.
        """
        self.detector_data = {}
        if detector_data is not None:
            self.detector_data.update(detector_data)
        if fname_pickle is not None:
            with open(fname_pickle, "rb") as picklefile:
                self.detector_data.update(pickle.load(picklefile))
        self.sample_rate = sample_rate
        self._radius = radius_deg
        self._get_pol_angles()
        self._get_pol_efficiency()

    def _get_pol_angles(self):
        """Get the detector polarization angles from the quaternions"""
        for detname, detdata in self.detector_data.items():
            if "pol_angle_deg" in detdata or "pol_angle_rad" in detdata:
                continue
            quat = detdata["quat"]
            theta, phi = qarray.to_position(quat)
            yrot = qarray.rotation(YAXIS, -theta)
            zrot = qarray.rotation(ZAXIS, -phi)
            rot = qarray.norm(qarray.mult(yrot, zrot))
            pol_rot = qarray.mult(rot, quat)
            pol_angle = qarray.to_angles(pol_rot)[2]
            detdata["pol_angle_rad"] = pol_angle
        return

    def _get_pol_efficiency(self):
        """Get the polarization efficiency from polarization leakage
        or vice versa
        """
        for detname, detdata in self.detector_data.items():
            if "pol_leakage" in detdata and "pol_efficiency" not in detdata:
                # Derive efficiency from leakage
                epsilon = detdata["pol_leakage"]
                eta = (1 - epsilon) / (1 + epsilon)
                detdata["pol_efficiency"] = eta
            elif "pol_leakage" not in detdata and "pol_efficiency" in detdata:
                # Derive leakage from efficiency
                eta = detdata["pol_effiency"]
                epsilon = (1 - eta) / (1 + eta)
                detdata["pol_leakage"] = epsilon
            elif "pol_leakage" not in detdata and "pol_efficiency" not in detdata:
                # Assume a perfectly polarized detector
                detdata["pol_efficiency"] = 1
                detdata["pol_leakage"] = 0
            else:
                # Check that efficiency and leakage are consistent
                epsilon = detdata["pol_leakage"]
                eta = detdata["pol_efficiency"]
                np.testing.assert_almost_equal(
                    eta,
                    (1 + epsilon) / (1 - epsilon),
                    err_msg="inconsistent polarization leakage and efficiency",
                )
        return

    def __contains__(self, key):
        return key in self.detector_data

    def __getitem__(self, key):
        return self.detector_data[key]

    def __setitem__(self, key, value):
        self.detector_data[key] = value

    def reset_properties(self):
        """Clear automatic properties so they will be re-generated"""
        self._detweights = None
        self._radius = None
        self._detquats = None
        self._noise = None

    @property
    def detweights(self):
        """Return the inverse noise variance weights [K_CMB^-2]"""
        if self._detweights is None:
            self._detweights = {}
            for detname, detdata in self.detector_data.items():
                net = detdata["NET"]
                if "fsample" in detdata:
                    fsample = detdata["fsample"]
                else:
                    fsample = self.sample_rate
                detweight = 1.0 / (fsample * net ** 2)
                self._detweights[detname] = detweight
        return self._detweights

    @property
    def radius(self):
        """The focal plane radius in degrees"""
        if self._radius is None:
            # Find the largest distance from the bore sight
            ZAXIS = np.array([0, 0, 1])
            cosangs = []
            for detname, detdata in self.detector_data.items():
                quat = detdata["quat"]
                vec = qarray.rotate(quat, ZAXIS)
                cosangs.append(np.dot(ZAXIS, vec))
            mincos = np.amin(cosangs)
            self._radius = np.degrees(np.arccos(mincos))
            # Add a very small margin to avoid numeric issues
            # in the atmospheric simulation
            self._radius *= 1.001
        return self._radius

    @property
    def detquats(self):
        if self._detquats is None:
            self._detquats = {}
            for detname, detdata in self.detector_data.items():
                self._detquats[detname] = detdata["quat"]
        return self._detquats

    @property
    def noise(self):
        if self._noise is None:
            detectors = sorted(self.detector_data.keys())
            fmin = {}
            fknee = {}
            alpha = {}
            NET = {}
            rates = {}
            for detname in detectors:
                detdata = self.detector_data[detname]
                if "fsample" in detdata:
                    rates[detname] = detdata["fsample"]
                else:
                    rates[detname] = self.sample_rate
                fmin[detname] = detdata["fmin"]
                fknee[detname] = detdata["fknee"]
                alpha[detname] = detdata["alpha"]
                NET[detname] = detdata["NET"]
            self._noise = AnalyticNoise(
                rate=rates,
                fmin=fmin,
                detectors=detectors,
                fknee=fknee,
                alpha=alpha,
                NET=NET,
            )
        return self._noise

    def __repr__(self):
        value = (
            "(Focalplane : {} detectors, sample_rate = {} Hz, radius = {} deg, "
            "detectors = ("
            "".format(len(self.detector_data), self.sample_rate, self.radius)
        )
        for detector_name, detector_data in self.detector_data.items():
            value += "{}, ".format(detector_name)
        value += "))"
        return value


class Telescope(object):
    def __init__(self, name, focalplane=None, site=None):
        self.name = name
        self.id = name2id(name)
        self.focalplane = focalplane
        self.site = site

    def __repr__(self):
        value = "(Telescope '{}' : ID = {}, Site = {}, Focalplane = {}" "".format(
            self.name, self.id, self.site, self.focalplane
        )
        return value



# Copyright (c) 2019-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import argparse
import os

import numpy as np

from ..timing import function_timer, Timer
from ..utils import Logger, Environment

from ..map import DistPixels
from ..todmap import OpMadam, OpPointingHpix


def add_pointing_args(parser):
    """Add the pointing arguments"""

    # `nside` may already be added
    try:
        parser.add_argument(
            "--nside", required=False, default=512, type=np.int, help="Healpix NSIDE"
        )
    except argparse.ArgumentError:
        pass
    parser.add_argument(
        "--nside-submap",
        required=False,
        default=16,
        type=np.int,
        help="Number of sub pixels is 12 * nside-submap ** 2",
    )
    # `coord` may already be added
    try:
        parser.add_argument(
            "--coord", required=False, default="C", help="Sky coordinate system [C,E,G]"
        )
    except argparse.ArgumentError:
        pass

    parser.add_argument(
        "--single-precision-pointing",
        required=False,
        action="store_true",
        help="Use single precision for pointing in memory.",
        dest="single_precision_pointing",
    )
    parser.add_argument(
        "--no-single-precision-pointing",
        required=False,
        action="store_false",
        help="Use double precision for pointing in memory.",
        dest="single_precision_pointing",
    )
    parser.set_defaults(single_precision_pointing=False)

    # Common flag mask may already be added
    try:
        parser.add_argument(
            "--common-flag-mask",
            required=False,
            default=1,
            type=np.uint8,
            help="Common flag mask",
        )
    except argparse.ArgumentError:
        pass
    # `flush` may already be added
    try:
        parser.add_argument(
            "--flush",
            required=False,
            default=False,
            action="store_true",
            help="Flush every print statement.",
        )
    except argparse.ArgumentError:
        pass
    return


@function_timer
def expand_pointing(args, comm, data):
    """Expand boresight pointing to every detector."""
    log = Logger.get()
    timer = Timer()
    timer.start()

    if comm.world_rank == 0:
        log.info("Expanding pointing")

    pointing = OpPointingHpix(
        nside=args.nside,
        nest=True,
        mode="IQU",
        single_precision=args.single_precision_pointing,
        nside_submap=args.nside_submap,
    )

    pointing.exec(data)

    if comm.comm_world is not None:
        comm.comm_world.barrier()
    if comm.world_rank == 0:
        timer.report_clear("Pointing generation")

    return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import numpy as np

from .mpi import MPITestCase

from .. import qarray as qa


class QarrayTest(MPITestCase):
    def setUp(self):
        # data
        self.q1 = np.array([0.50487417, 0.61426059, 0.60118994, 0.07972857])
        self.q1inv = np.array([-0.50487417, -0.61426059, -0.60118994, 0.07972857])
        self.q2 = np.array([0.43561544, 0.33647027, 0.40417115, 0.73052901])
        self.qtonormalize = np.array([[1.0, 2, 3, 4], [2, 3, 4, 5]])
        self.qnormalized = np.array(
            [
                [0.18257419, 0.36514837, 0.54772256, 0.73029674],
                [0.27216553, 0.40824829, 0.54433105, 0.68041382],
            ]
        )
        self.vec = np.array([0.57734543, 0.30271255, 0.75831218])
        self.vec2 = np.array(
            [[0.57734543, 8.30271255, 5.75831218], [1.57734543, 3.30271255, 0.75831218]]
        )
        self.qeasy = np.array([[0.3, 0.3, 0.1, 0.9], [0.3, 0.3, 0.1, 0.9]])
        # results from Quaternion
        # CHANGED SIGN TO COMPLY WITH THE ONLINE QUATERNION CALCULATOR
        self.mult_result = -1 * np.array(
            [-0.44954009, -0.53339352, -0.37370443, 0.61135101]
        )
        self.rot_by_q1 = np.array([0.4176698, 0.84203849, 0.34135482])
        self.rot_by_q2 = np.array([0.8077876, 0.3227185, 0.49328689])

    def test_inv(self):
        np.testing.assert_array_almost_equal(qa.inv(self.q1), self.q1inv)
        return

    def test_norm(self):
        np.testing.assert_array_almost_equal(
            qa.norm(self.q1), self.q1 / np.linalg.norm(self.q1)
        )
        np.testing.assert_array_almost_equal(
            qa.norm(self.qtonormalize), self.qnormalized
        )
        return

    def test_mult_onequaternion(self):
        my_mult_result = qa.mult(self.q1, self.q2)
        np.testing.assert_array_almost_equal(my_mult_result, self.mult_result)
        return

    def test_mult_qarray(self):
        dim = (3, 1)
        qarray1 = np.tile(self.q1, dim)
        qarray2 = np.tile(self.q2, dim)
        my_mult_result = qa.mult(qarray1, qarray2)
        np.testing.assert_array_almost_equal(
            my_mult_result, np.tile(self.mult_result, dim)
        )

        check = qa.mult(self.q1, self.q2)
        res = qa.mult(np.tile(self.q1, 10).reshape((-1, 4)), self.q2)
        np.testing.assert_array_almost_equal(res, np.tile(check, 10).reshape((-1, 4)))

        nulquat = np.array([0.0, 0.0, 0.0, 1.0])
        check = qa.mult(self.q1, nulquat)
        res = qa.mult(np.tile(self.q1, 10).reshape((-1, 4)), nulquat)
        np.testing.assert_array_almost_equal(res, np.tile(check, 10).reshape((-1, 4)))
        return

    def test_rotate_onequaternion(self):
        my_rot_result = qa.rotate(self.q1, self.vec)
        np.testing.assert_array_almost_equal(my_rot_result, self.rot_by_q1)
        return

    def test_rotate_qarray(self):
        my_rot_result = qa.rotate(np.vstack([self.q1, self.q2]), self.vec)
        np.testing.assert_array_almost_equal(
            my_rot_result, np.vstack([self.rot_by_q1, self.rot_by_q2]).reshape((2, 3))
        )

        zaxis = np.array([0.0, 0.0, 1.0])

        nsamp = 1000
        theta = (1.0 / np.pi) * np.arange(nsamp, dtype=np.float64)
        phi = (10.0 / (2.0 * np.pi)) * np.arange(nsamp, dtype=np.float64)
        pa = np.zeros(nsamp, dtype=np.float64)

        quats = qa.from_angles(theta, phi, pa)

        check = np.zeros((nsamp, 3), dtype=np.float64)
        for i in range(nsamp):
            check[i, :] = qa.rotate(quats[i], zaxis)

        dir = qa.rotate(quats, zaxis)

        np.testing.assert_array_almost_equal(dir, check)
        return

    def test_slerp(self):
        q = qa.norm(np.array([[2.0, 3, 4, 5], [6.0, 7, 8, 9]]))
        time = np.array([0.0, 9])
        targettime = np.array([0, 3, 4.5, 9])
        q_interp = qa.slerp(targettime, time, q)
        self.assertEqual(len(q_interp), 4)
        np.testing.assert_array_almost_equal(q_interp[0], q[0])
        np.testing.assert_array_almost_equal(q_interp[-1], q[-1])
        np.testing.assert_array_almost_equal(
            q_interp[1], qa.norm(q[0] * 2 / 3 + q[1] / 3), decimal=4
        )
        np.testing.assert_array_almost_equal(
            q_interp[2], qa.norm((q[0] + q[1]) / 2), decimal=4
        )
        return

    def test_rotation(self):
        np.testing.assert_array_almost_equal(
            qa.rotation(np.array([0.0, 0.0, 1.0]), np.radians(30)),
            np.array([0, 0, np.sin(np.radians(15)), np.cos(np.radians(15))]),
        )
        return

    def test_toaxisangle(self):
        axis = np.array([0.0, 0.0, 1.0])
        angle = np.radians(30.0)
        q = np.array([0, 0, np.sin(np.radians(15)), np.cos(np.radians(15))])
        qaxis, qangle = qa.to_axisangle(q)
        np.testing.assert_array_almost_equal(axis, qaxis)
        self.assertAlmostEqual(angle, qangle)
        return

    def test_exp(self):
        # Exponential test from:
        #     http://world.std.com/~sweetser/java/qcalc/qcalc.html
        np.testing.assert_array_almost_equal(
            qa.exp(self.qeasy),
            np.array(
                [
                    [0.71473568, 0.71473568, 0.23824523, 2.22961712],
                    [0.71473568, 0.71473568, 0.23824523, 2.22961712],
                ]
            ),
        )
        return

    def test_ln(self):
        # Log test from: http://world.std.com/~sweetser/java/qcalc/qcalc.html
        np.testing.assert_array_almost_equal(
            qa.ln(self.qeasy),
            np.array(
                [
                    [0.31041794, 0.31041794, 0.10347265, 0.0],
                    [0.31041794, 0.31041794, 0.10347265, 0.0],
                ]
            ),
        )
        return

    def test_pow(self):
        # Pow test from: http://world.std.com/~sweetser/java/qcalc/qcalc.html
        np.testing.assert_array_almost_equal(
            qa.pow(self.qeasy, 3.0),
            np.array([[0.672, 0.672, 0.224, 0.216], [0.672, 0.672, 0.224, 0.216]]),
        )
        np.testing.assert_array_almost_equal(
            qa.pow(self.qeasy, 0.1),
            np.array(
                [
                    [0.03103127, 0.03103127, 0.01034376, 0.99898305],
                    [0.03103127, 0.03103127, 0.01034376, 0.99898305],
                ]
            ),
        )
        return

    def test_torotmat(self):
        # Rotmat test from Quaternion
        np.testing.assert_array_almost_equal(
            qa.to_rotmat(self.qeasy[0]),
            np.array(
                [
                    [8.00000000e-01, -2.77555756e-17, 6.00000000e-01],
                    [3.60000000e-01, 8.00000000e-01, -4.80000000e-01],
                    [-4.80000000e-01, 6.00000000e-01, 6.40000000e-01],
                ]
            ),
        )
        return

    def test_fromrotmat(self):
        np.testing.assert_array_almost_equal(
            self.qeasy[0], qa.from_rotmat(qa.to_rotmat(self.qeasy[0]))
        )
        return

    def test_fromvectors(self):
        angle = np.radians(30)

        v1 = np.array([1.0, 0.0, 0.0])
        v2 = np.array([np.cos(angle), np.sin(angle), 0])
        np.testing.assert_array_almost_equal(
            qa.from_vectors(v1, v2),
            np.array([0, 0, np.sin(np.radians(15)), np.cos(np.radians(15))]),
        )

        nulquat = np.array([0.0, 0.0, 0.0, 1.0])
        zaxis = np.array([0.0, 0.0, 1.0])
        check = qa.from_vectors(zaxis, zaxis)
        np.testing.assert_array_almost_equal(check, nulquat)

        q = qa.from_vectors(
            np.tile(v1, 3).reshape(-1, 3), np.tile(v2, 3).reshape(-1, 3)
        )

        comp = np.tile(
            np.array([0, 0, np.sin(np.radians(15)), np.cos(np.radians(15))]), 3
        ).reshape(-1, 4)

        np.testing.assert_array_almost_equal(q, comp)
        return

    def test_fp(self):
        xaxis = np.array([1, 0, 0], dtype=np.float64)
        zaxis = np.array([0, 0, 1], dtype=np.float64)

        radius = np.deg2rad(10.0)

        for pos in range(6):
            posang = np.deg2rad(pos * 60.0)
            posrot = qa.rotation(zaxis, posang + np.pi / 2.0)
            radrot = qa.rotation(xaxis, radius)
            detrot = qa.mult(posrot, radrot)

            detdir = qa.rotate(detrot, zaxis)

            check = np.dot(detdir, zaxis)

            np.testing.assert_almost_equal(check, np.cos(radius))
        return

    def test_angles(self):
        ntheta = 5
        nphi = 5
        n = ntheta * nphi
        xaxis = np.array([1.0, 0.0, 0.0])
        zaxis = np.array([0.0, 0.0, 1.0])

        theta = np.zeros(n, dtype=np.float64)
        phi = np.zeros(n, dtype=np.float64)
        pa = np.zeros(n, dtype=np.float64)

        # Healpix convention

        for i in range(ntheta):
            for j in range(nphi):
                theta[i * nphi + j] = (0.5 + i) * np.pi / float(ntheta)
                phi[i * nphi + j] = j * 2.0 * np.pi / float(nphi)
                pa[i * nphi + j] = j * 2.0 * np.pi / float(nphi) - np.pi

        quat = qa.from_angles(theta, phi, pa, IAU=False)

        dir = qa.rotate(quat, np.tile(zaxis, n).reshape((n, 3)))
        orient = qa.rotate(quat, np.tile(xaxis, n).reshape((n, 3)))

        np.testing.assert_array_almost_equal(
            np.pi / 2 - np.arcsin(dir[:, 2]), theta, decimal=4
        )

        check = np.arctan2(dir[:, 1], dir[:, 0])

        check[check < 0.0] += 2.0 * np.pi
        check[check > 2.0 * np.pi] -= 2.0 * np.pi
        check[(np.absolute(check) < 2.0e-16)] = 0.0
        check[(np.absolute(check - 2.0 * np.pi) < 2.0e-16)] = 0.0

        np.testing.assert_array_almost_equal(check, phi, decimal=6)

        check = np.arctan2(
            orient[:, 0] * dir[:, 1] - orient[:, 1] * dir[:, 0],
            -(orient[:, 0] * dir[:, 2] * dir[:, 0])
            - (orient[:, 1] * dir[:, 2] * dir[:, 1])
            + (orient[:, 2] * (dir[:, 0] * dir[:, 0] + dir[:, 1] * dir[:, 1])),
        )

        np.testing.assert_array_almost_equal(check, pa, decimal=6)

        check_theta, check_phi, check_pa = qa.to_angles(quat, IAU=False)

        np.testing.assert_array_almost_equal(check_theta, theta, decimal=4)

        check_phi[(np.absolute(check_phi) < 2.0e-6)] = 0.0
        check_phi[(np.absolute(check_phi - 2.0 * np.pi) < 2.0e-6)] = 0.0

        np.testing.assert_array_almost_equal(check_phi, phi, decimal=4)

        np.testing.assert_array_almost_equal(check_pa, pa, decimal=4)

        # IAU convention

        for i in range(ntheta):
            for j in range(nphi):
                theta[i * nphi + j] = (0.5 + i) * np.pi / float(ntheta)
                phi[i * nphi + j] = j * 2.0 * np.pi / float(nphi)
                pa[i * nphi + j] = -j * 2.0 * np.pi / float(nphi) - np.pi

        quat = qa.from_angles(theta, phi, pa, IAU=True)

        dir = qa.rotate(quat, np.tile(zaxis, n).reshape((n, 3)))
        orient = qa.rotate(quat, np.tile(xaxis, n).reshape((n, 3)))

        np.testing.assert_array_almost_equal(
            np.pi / 2 - np.arcsin(dir[:, 2]), theta, decimal=4
        )

        check = np.arctan2(dir[:, 1], dir[:, 0])

        check[check < 0.0] += 2.0 * np.pi
        check[check > 2.0 * np.pi] -= 2.0 * np.pi
        check[(np.absolute(check) < 2.0e-16)] = 0.0
        check[(np.absolute(check - 2.0 * np.pi) < 2.0e-16)] = 0.0

        np.testing.assert_array_almost_equal(check, phi, decimal=6)

        check = -(
            np.arctan2(
                orient[:, 0] * dir[:, 1] - orient[:, 1] * dir[:, 0],
                -(orient[:, 0] * dir[:, 2] * dir[:, 0])
                - (orient[:, 1] * dir[:, 2] * dir[:, 1])
                + (orient[:, 2] * (dir[:, 0] * dir[:, 0] + dir[:, 1] * dir[:, 1])),
            )
        )

        check[check < -np.pi] += 2.0 * np.pi
        check[check > np.pi] -= 2.0 * np.pi

        comp_pa = np.copy(pa)
        comp_pa[comp_pa < -np.pi] += 2.0 * np.pi
        comp_pa[comp_pa > np.pi] -= 2.0 * np.pi

        np.testing.assert_array_almost_equal(check, comp_pa, decimal=6)

        check_theta, check_phi, check_pa = qa.to_angles(quat, IAU=True)

        np.testing.assert_array_almost_equal(check_theta, theta, decimal=4)

        check_phi[(np.absolute(check_phi) < 2.0e-6)] = 0.0
        check_phi[(np.absolute(check_phi - 2.0 * np.pi) < 2.0e-6)] = 0.0

        np.testing.assert_array_almost_equal(check_phi, phi, decimal=4)

        np.testing.assert_array_almost_equal(check_pa, comp_pa, decimal=4)

        # to_position test

        check_theta, check_phi = qa.to_position(quat)
        check_theta2, check_phi2, check_pa = qa.to_angles(quat, IAU=False)

        np.testing.assert_array_almost_equal(check_theta, check_theta2, decimal=4)
        np.testing.assert_array_almost_equal(check_phi, check_phi2, decimal=4)
        return

    def test_depths(self):
        # Verify that qarray methods preserve the depths of their inputs
        np.testing.assert_equal(qa.mult(self.q1, self.q2).shape, (4,))
        np.testing.assert_equal(qa.mult(np.atleast_2d(self.q1), self.q2).shape, (1, 4))

        np.testing.assert_equal(qa.inv(self.q1).shape, (4,))
        np.testing.assert_equal(qa.inv(np.atleast_2d(self.q1)).shape, (1, 4))

        np.testing.assert_equal(np.shape(qa.amplitude(self.q1)), ())
        np.testing.assert_equal(np.shape(qa.amplitude(np.atleast_2d(self.q1))), (1,))

        np.testing.assert_equal(qa.norm(self.q1).shape, (4,))
        np.testing.assert_equal(qa.norm(np.atleast_2d(self.q1)).shape, (1, 4))

        np.testing.assert_equal(qa.rotate(self.q1, self.vec).shape, (3,))
        np.testing.assert_equal(
            qa.rotate(np.atleast_2d(self.q1), self.vec).shape, (1, 3)
        )
        np.testing.assert_equal(
            qa.rotate(self.q1, np.atleast_2d(self.vec)).shape, (1, 3)
        )

        q = qa.norm(np.array([[2.0, 3, 4, 5], [6.0, 7, 8, 9]]))
        time = np.array([0.0, 9])
        np.testing.assert_equal(np.shape(qa.slerp(0, time, q)), (4,))
        np.testing.assert_equal(np.shape(qa.slerp([0], time, q)), (1, 4))

        np.testing.assert_equal(qa.exp(self.q1).shape, (4,))
        np.testing.assert_equal(qa.exp(np.atleast_2d(self.q1)).shape, (1, 4))

        np.testing.assert_equal(qa.ln(self.q1).shape, (4,))
        np.testing.assert_equal(qa.ln(np.atleast_2d(self.q1)).shape, (1, 4))

        np.testing.assert_equal(qa.pow(self.q1, 2).shape, (4,))
        np.testing.assert_equal(qa.pow(np.atleast_2d(self.q1), 2).shape, (1, 4))
        np.testing.assert_equal(qa.pow(self.q1, [2]).shape, (1, 4))
        np.testing.assert_equal(qa.pow(np.atleast_2d(self.q1), [2]).shape, (1, 4))

        np.testing.assert_equal(qa.rotation([0, 0, 1], np.pi).shape, (4,))
        np.testing.assert_equal(qa.rotation([[0, 0, 1]], np.pi).shape, (1, 4))
        np.testing.assert_equal(qa.rotation([0, 0, 1], [np.pi]).shape, (1, 4))
        np.testing.assert_equal(qa.rotation([[0, 0, 1]], [np.pi]).shape, (1, 4))

        ret1 = qa.to_axisangle(self.q1)
        ret2 = qa.to_axisangle(np.atleast_2d(self.q1))
        np.testing.assert_equal(ret1[0].shape, (3,))
        np.testing.assert_equal(np.shape(ret1[1]), ())
        np.testing.assert_equal(ret2[0].shape, (1, 3))
        np.testing.assert_equal(np.shape(ret2[1]), (1,))

        np.testing.assert_equal(np.shape(qa.from_angles(0, 0, 0)), (4,))
        np.testing.assert_equal(np.shape(qa.from_angles([0], 0, 0)), (1, 4))
        np.testing.assert_equal(np.shape(qa.from_angles([0], [0], [0])), (1, 4))

        ret1 = qa.to_angles(self.q1)
        ret2 = qa.to_angles(np.atleast_2d(self.q1))
        np.testing.assert_equal(np.shape(ret1[0]), ())
        np.testing.assert_equal(np.shape(ret1[1]), ())
        np.testing.assert_equal(np.shape(ret1[2]), ())
        np.testing.assert_equal(np.shape(ret2[0]), (1,))
        np.testing.assert_equal(np.shape(ret2[1]), (1,))
        np.testing.assert_equal(np.shape(ret2[2]), (1,))
        return



# Copyright (c) 2019-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import argparse
from datetime import datetime
import os

import numpy as np

from ..mpi import get_world, Comm
from ..timing import function_timer, Timer
from ..utils import Logger, Environment


def add_dist_args(parser):
    parser.add_argument(
        "--group-size",
        required=False,
        type=np.int,
        help="Size of a process group assigned to an observation",
    )
    try:
        parser.add_argument(
            "--day-maps",
            required=False,
            action="store_true",
            help="Enable daily maps",
            dest="do_daymaps",
        )
        parser.add_argument(
            "--no-day-maps",
            required=False,
            action="store_false",
            help="Disable daily maps",
            dest="do_daymaps",
        )
        parser.set_defaults(do_daymaps=False)
    except argparse.ArgumentError:
        # day-maps was already defined
        pass

    try:
        parser.add_argument(
            "--season-maps",
            required=False,
            action="store_true",
            help="Enable season maps",
            dest="do_seasonmaps",
        )
        parser.add_argument(
            "--no-season-maps",
            required=False,
            action="store_false",
            help="Disable season maps",
            dest="do_seasonmaps",
        )
        parser.set_defaults(do_seasonmaps=False)
    except argparse.ArgumentError:
        # season-maps was already defined
        pass

    return


def get_comm():
    log = Logger.get()
    env = Environment.get()
    mpiworld, procs, rank = get_world()
    if rank == 0:
        print(env)
    if mpiworld is None:
        log.info("Running serially with one process at {}".format(str(datetime.now())))
    else:
        if rank == 0:
            log.info(
                "Running with {} processes at {}".format(procs, str(datetime.now()))
            )

    # This is the 2-level toast communicator.  By default,
    # there is just one group which spans MPI_COMM_WORLD.
    comm = Comm(world=mpiworld)
    return mpiworld, procs, rank, comm


@function_timer
def get_time_communicators(args, comm, data):
    """Split the world communicator by time."""
    time_comms = [("all", comm.comm_world)]
    if comm.comm_world is None:
        return time_comms

    # A process will only have data for one season and one day.  If more
    # than one season is observed, we split the communicator to make
    # season maps.

    if args.do_seasonmaps:
        my_season = data.obs[0]["season"]
        seasons = np.array(comm.comm_world.allgather(my_season))
        do_seasons = np.any(seasons != my_season)
        if do_seasons:
            season_comm = comm.comm_world.Split(my_season, comm.world_rank)
            time_comms.append((str(my_season), season_comm))

    if args.do_daymaps:
        # Split the communicator to make daily maps.  We could easily split
        # by month as well

        my_day = int(data.obs[0]["MJD"])
        my_date = data.obs[0]["date"]
        days = np.array(comm.comm_world.allgather(my_day))
        do_days = np.any(days != my_day)
        if do_days:
            day_comm = comm.comm_world.Split(my_day, comm.world_rank)
            time_comms.append((my_date, day_comm))

    return time_comms



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import re

from ..op import Operator

from ..rng import random

from ..utils import Logger

from ..timing import function_timer


class OpGainScrambler(Operator):
    """Apply random gain errors to detector data.

    This operator draws random gain errors from a given distribution and
    applies them to the specified detectors.

    Args:
        center (float):  Gain distribution center.
        sigma (float):  Gain distribution width.
        pattern (str):  Regex pattern to match against detector names.
            Only detectors that match the pattern are scrambled.
        name (str):  Name of the output signal cache object will be
            <name_in>_<detector>.  If the object exists, it is used as
            input.  Otherwise signal is read using the tod read method.
        realization (int): if simulating multiple realizations, the
            realization index.
        component (int): the component index to use for this noise
            simulation.

    """

    def __init__(
        self,
        center=1,
        sigma=1e-3,
        pattern=r".*",
        name=None,
        realization=0,
        component=234567,
    ):
        self._center = center
        self._sigma = sigma
        self._pattern = pattern
        self._name = name
        self._realization = realization
        self._component = component

        # Call the parent class constructor.
        super().__init__()

    @function_timer
    def exec(self, data):
        """Scramble the gains.

        Args:
            data (toast.Data): The distributed data.

        """
        log = Logger.get()
        for obs in data.obs:
            tod = obs["tod"]
            comm = tod.mpicomm
            rank = 0
            if comm is not None:
                rank = comm.rank

            obsindx = 0
            if "id" in obs:
                obsindx = obs["id"]
            else:
                if rank == 0:
                    log.warning("observation ID is not set, using zero!")

            telescope = 0
            if "telescope" in obs:
                telescope = obs["telescope_id"]

            pat = re.compile(self._pattern)

            for det in tod.local_dets:
                # Test the detector pattern
                if pat.match(det) is None:
                    continue

                detindx = tod.detindx[det]

                # Cache the output signal
                ref = tod.local_signal(det, self._name)

                # key1 = realization * 2^32 + telescope * 2^16 + component
                # key2 = obsindx * 2^32 + detindx
                # counter1 = currently unused (0)
                # counter2 = currently unused (0)

                key1 = (
                    self._realization * 4294967296 + telescope * 65536 + self._component
                )
                key2 = obsindx * 4294967296 + detindx
                counter1 = 0
                counter2 = 0

                rngdata = random(
                    1,
                    sampler="gaussian",
                    key=(key1, key2),
                    counter=(counter1, counter2),
                )

                gain = self._center + rngdata[0] * self._sigma
                ref *= gain

                del ref

        return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

from .mpi import MPITestCase

import os

import numpy as np

import numpy.testing as nt

import healpy as hp

from ..tod import AnalyticNoise, OpSimNoise
from ..todmap import TODSatellite, OpPointingHpix, OpAccumDiag
from ..todmap.todmap_math import cov_accum_diag
from ..map import DistPixels, covariance_invert, covariance_rcond, covariance_multiply

from ._helpers import (
    create_outdir,
    create_distdata,
    boresight_focalplane,
    uniform_chunks,
)


class CovarianceTest(MPITestCase):
    def setUp(self):
        fixture_name = os.path.splitext(os.path.basename(__file__))[0]
        self.outdir = create_outdir(self.comm, fixture_name)

        # Create one observation per group, and each observation will have
        # a fixed number of detectors and one chunk per process.

        self.data = create_distdata(self.comm, obs_per_group=1)

        self.ndet = 1
        self.rate = 40.0
        self.hwprpm = 50

        # Create detectors
        (
            dnames,
            dquat,
            depsilon,
            drate,
            dnet,
            dfmin,
            dfknee,
            dalpha,
        ) = boresight_focalplane(self.ndet, samplerate=self.rate, net=7.0)

        # Samples per observation
        self.totsamp = 240000

        # Pixelization

        self.sim_nside = 32
        self.sim_npix = 12 * self.sim_nside ** 2

        self.map_nside = 32
        self.map_npix = 12 * self.map_nside ** 2

        # Scan strategy

        self.spinperiod = 10.0
        self.spinangle = 30.0
        self.precperiod = 50.0
        self.precangle = 65.0

        # One chunk per process
        chunks = uniform_chunks(self.totsamp, nchunk=self.data.comm.group_size)

        # Populate the single observation per group

        tod = TODSatellite(
            self.data.comm.comm_group,
            dquat,
            self.totsamp,
            detranks=1,
            firsttime=0.0,
            rate=self.rate,
            spinperiod=self.spinperiod,
            spinangle=self.spinangle,
            precperiod=self.precperiod,
            precangle=self.precangle,
            sampsizes=chunks,
            hwprpm=self.hwprpm,
        )

        tod.set_prec_axis()

        nse = AnalyticNoise(
            rate=drate,
            fmin=dfmin,
            detectors=dnames,
            fknee=dfknee,
            alpha=dalpha,
            NET=dnet,
        )

        self.data.obs[0]["tod"] = tod
        self.data.obs[0]["noise"] = nse

    def tearDown(self):
        del self.data

    def test_accum(self):
        nsm = 2
        npix = 3
        nnz = 4
        block = int(nnz * (nnz + 1) / 2)
        scale = 2.0
        nsamp = nsm * npix

        fake = DistPixels(
            None,
            comm=self.data.comm.comm_world,
            npix=nsm * npix,
            nnz=nnz,
            dtype=np.float64,
            npix_submap=npix,
            local_submaps=np.arange(nsm, dtype=np.int64),
        )
        check = fake.duplicate()

        hits = DistPixels(
            None,
            comm=self.data.comm.comm_world,
            npix=nsm * npix,
            nnz=1,
            dtype=np.int64,
            npix_submap=npix,
            local_submaps=np.arange(nsm, dtype=np.int64),
        )
        checkhits = hits.duplicate()

        invn = DistPixels(
            None,
            comm=self.data.comm.comm_world,
            npix=nsm * npix,
            nnz=block,
            dtype=np.float64,
            npix_submap=npix,
            local_submaps=np.arange(nsm, dtype=np.int64),
        )
        checkinvn = invn.duplicate()

        sm = np.zeros(nsamp, dtype=np.int64)
        pix = np.zeros(nsamp, dtype=np.int64)
        wt = np.zeros(nsamp * nnz, dtype=np.float64)

        for i in range(nsamp):
            sm[i] = i % nsm
            pix[i] = i % npix
            for k in range(nnz):
                wt[i * nnz + k] = float(k + 1)

        signal = np.random.normal(size=nsamp)
        #
        # print(
        #     nsm,
        #     npix,
        #     nnz,
        #     sm.dtype,
        #     pix.dtype,
        #     wt.dtype,
        #     scale,
        #     signal.dtype,
        #     invn.flatdata.dtype,
        #     hits.flatdata.dtype,
        #     fake.flatdata.dtype,
        #     flush=True,
        # )

        cov_accum_diag(
            nsm,
            npix,
            nnz,
            sm,
            pix,
            wt,
            scale,
            signal,
            invn.flatdata,
            hits.flatdata,
            fake.flatdata,
        )

        for i in range(nsamp):
            checkhits.data[sm[i], pix[i], 0] += 1
            off = 0
            for j in range(nnz):
                check.data[sm[i], pix[i], j] += scale * signal[i] * wt[i * nnz + j]
                for k in range(j, nnz):
                    checkinvn.data[sm[i], pix[i], off] += (
                        scale * wt[i * nnz + j] * wt[i * nnz + k]
                    )
                    off += 1

        # for i in range(nsamp):
        #     print("{}: {} {}".format(i, checkhits.data[sm[i], pix[i], 0], hits.data[sm[i], pix[i], 0]))
        #     off = 0
        #     for j in range(nnz):
        #         print("  {}:  {}  {}".format(j, check.data[sm[i], pix[i], j], fake.data[sm[i], pix[i], j]))
        #         for k in range(j, nnz):
        #             print("    {}:  {}  {}".format(off, checkinvn.data[sm[i], pix[i], off], invn.data[sm[i], pix[i], off]))
        #             off += 1

        nt.assert_equal(hits.data, checkhits.data)
        nt.assert_almost_equal(fake.data, check.data)
        nt.assert_almost_equal(invn.data, checkinvn.data)

        return

    def test_invert(self):
        nsm = 2
        npix = 3
        nnz = 4
        scale = 2.0
        nsamp = nsm * npix
        nelem = int(nnz * (nnz + 1) / 2)
        threshold = 1.0e-6

        invn = DistPixels(
            None,
            comm=self.data.comm.comm_world,
            npix=nsm * npix,
            nnz=nelem,
            dtype=np.float64,
            npix_submap=npix,
            local_submaps=np.arange(nsm, dtype=np.int64),
        )
        check = invn.duplicate()

        rowdata = 10.0 * np.arange(nnz, 0, -1)

        for i in range(nsm):
            for j in range(npix):
                off = 0
                for k in range(nnz):
                    for m in range(k, nnz):
                        invn.data[i, j, off] = rowdata[m - k]
                        check.data[i, j, off] = invn.data[i, j, off]
                        off += 1

        # invert twice
        covariance_invert(invn, threshold)
        covariance_invert(invn, threshold)

        # for i in range(nsm):
        #     for j in range(npix):
        #         off = 0
        #         print("sm {}, pix {}:".format(i, j))
        #         for k in range(nnz):
        #             for m in range(k, nnz):
        #                 print("  {} {}".format(fakedata[i,j,off], checkdata[i,j,off]))
        #                 off += 1

        nt.assert_almost_equal(invn.data, check.data)
        return

    def test_invnpp(self):
        op = OpSimNoise(realization=0)
        op.exec(self.data)

        # make a simple pointing matrix
        pointing = OpPointingHpix(nside=self.map_nside, nest=True, mode="IQU")
        pointing.exec(self.data)

        # construct a distributed map to store the covariance and hits

        invnpp = DistPixels(
            self.data,
            nnz=6,
            dtype=np.float64,
        )
        invnpp.data.fill(0.0)

        hits = DistPixels(
            self.data,
            nnz=1,
            dtype=np.int64,
        )
        hits.data.fill(0)

        # accumulate the inverse covariance.  Use detector weights
        # based on the analytic NET.

        tod = self.data.obs[0]["tod"]
        nse = self.data.obs[0]["noise"]
        detweights = {}
        for d in tod.local_dets:
            detweights[d] = 1.0 / (self.rate * nse.NET(d) ** 2)

        build_invnpp = OpAccumDiag(
            detweights=detweights, invnpp=invnpp, hits=hits, name="noise"
        )
        build_invnpp.exec(self.data)

        # for i in range(invnpp.data.shape[0]):
        #     for j in range(invnpp.data.shape[1]):
        #         print("sm {}, pix {}:  hits = {}".format(i, j, hits.data[i, j, 0]))
        #         for k in range(invnpp.data.shape[2]):
        #             print("  {}".format(invnpp.data[i, j, k]))

        invnpp.allreduce()
        hits.allreduce()

        # invert it
        check = invnpp.duplicate()
        covariance_invert(invnpp, 1.0e-14)
        covariance_invert(invnpp, 1.0e-14)

        # Matrices that failed the rcond test are set to zero
        nonzero = np.absolute(invnpp.data) > 1.0e-12
        if np.sum(nonzero) == 0:
            raise Exception("All matrices failed the rcond test.")

        nt.assert_almost_equal(invnpp.data[nonzero], check.data[nonzero])

        return

    def test_distpix_init(self):
        # make a simple pointing matrix
        pointing = OpPointingHpix(nside=self.map_nside, nest=True, mode="IQU")
        pointing.exec(self.data)

        # construct a distributed map to store the covariance and hits

        invnpp = DistPixels(
            self.data,
            nnz=6,
            dtype=np.float64,
        )

        return

    def test_multiply(self):
        # make a simple pointing matrix
        pointing = OpPointingHpix(nside=self.map_nside, nest=True, mode="IQU")
        pointing.exec(self.data)

        # construct a distributed map to store the covariance and hits

        invnpp = DistPixels(
            self.data,
            nnz=6,
            dtype=np.float64,
        )

        hits = DistPixels(
            self.data,
            nnz=1,
            dtype=np.int64,
        )

        # accumulate the inverse covariance.  Use detector weights
        # based on the analytic NET.

        tod = self.data.obs[0]["tod"]
        nse = self.data.obs[0]["noise"]
        detweights = {}
        for d in tod.local_dets:
            detweights[d] = 1.0 / (self.rate * nse.NET(d) ** 2)

        build_invnpp = OpAccumDiag(detweights=detweights, invnpp=invnpp, hits=hits)
        build_invnpp.exec(self.data)

        invnpp.allreduce()
        hits.allreduce()

        # invert it
        check = invnpp.duplicate()
        covariance_invert(invnpp, 1.0e-3)

        # multiply the two
        covariance_multiply(check, invnpp)

        # check that the multiplied matrices are unit matrices
        nsubmap, npix, nblock = check.data.shape
        nnz = int(((np.sqrt(8 * nblock) - 1) / 2) + 0.5)
        for i in range(nsubmap):
            for j in range(npix):
                if np.all(invnpp.data[i, j] == 0):
                    # Matrix failed to invert
                    continue
                off = 0
                for k in range(nnz):
                    for m in range(k, nnz):
                        if k == m:
                            nt.assert_almost_equal(check.data[i, j, off], 1.0)
                        else:
                            nt.assert_almost_equal(check.data[i, j, off], 0.0)
                        off += 1

        return

    def test_fitsio(self):
        rank = 0
        if self.comm is not None:
            rank = self.comm.rank
        # make a simple pointing matrix
        pointing = OpPointingHpix(nside=self.map_nside, nest=True, mode="IQU")
        pointing.exec(self.data)

        # construct a distributed map to store the covariance and hits

        invnpp = DistPixels(
            self.data,
            nnz=6,
            dtype=np.float64,
        )

        rcond = DistPixels(
            self.data,
            nnz=1,
            dtype=np.float64,
        )

        hits = DistPixels(
            self.data,
            nnz=1,
            dtype=np.int64,
        )

        # accumulate the inverse covariance.  Use detector weights
        # based on the analytic NET.

        tod = self.data.obs[0]["tod"]
        nse = self.data.obs[0]["noise"]
        detweights = {}
        for d in tod.local_dets:
            detweights[d] = 1.0 / (self.rate * nse.NET(d) ** 2)

        build_invnpp = OpAccumDiag(detweights=detweights, invnpp=invnpp, hits=hits)
        build_invnpp.exec(self.data)

        # self.assertTrue(False)

        invnpp.allreduce()
        hits.allreduce()

        # invert it

        covariance_invert(invnpp, 1.0e-3)
        rcond = covariance_rcond(invnpp)

        check = invnpp.duplicate()
        checkhits = hits.duplicate()
        checkrcond = rcond.duplicate()

        # write this out...

        subsum = [
            np.sum(invnpp.data[x, :, :]) for x in range(len(invnpp.local_submaps))
        ]

        outfile = os.path.join(self.outdir, "covtest.fits")
        if rank == 0:
            if os.path.isfile(outfile):
                os.remove(outfile)

        hitfile = os.path.join(self.outdir, "covtest_hits.fits")
        if rank == 0:
            if os.path.isfile(hitfile):
                os.remove(hitfile)

        rcondfile = os.path.join(self.outdir, "covtest_rcond.fits")
        if rank == 0:
            if os.path.isfile(rcondfile):
                os.remove(rcondfile)

        invnpp.write_healpix_fits(outfile)
        rcond.write_healpix_fits(rcondfile)
        hits.write_healpix_fits(hitfile)

        invnpp.data.fill(0.0)
        invnpp.read_healpix_fits(outfile)

        diffdata = invnpp.duplicate()
        diffdata.data -= check.data

        difffile = os.path.join(self.outdir, "readwrite_diff.fits")
        diffdata.write_healpix_fits(difffile)

        if rank == 0:
            import matplotlib.pyplot as plt

            dat = hp.read_map(outfile)
            outfile = "{}.png".format(outfile)
            hp.mollview(dat, xsize=int(1600))
            plt.savefig(outfile)
            plt.close()

            dat = hp.read_map(difffile)
            outfile = "{}.png".format(difffile)
            hp.mollview(dat, xsize=int(1600))
            plt.savefig(outfile)
            plt.close()

            dat = hp.read_map(hitfile)
            outfile = "{}.png".format(hitfile)
            hp.mollview(dat, xsize=int(1600))
            plt.savefig(outfile)
            plt.close()

            dat = hp.read_map(rcondfile)
            outfile = "{}.png".format(rcondfile)
            hp.mollview(dat, xsize=int(1600))
            plt.savefig(outfile)
            plt.close()

        rcond.data.fill(0.0)
        rcond.read_healpix_fits(rcondfile)

        nt.assert_almost_equal(rcond.data, checkrcond.data, decimal=6)
        # nt.assert_almost_equal(invnpp.data, checkdata, decimal=6)

        hits.data.fill(0)
        hits.read_healpix_fits(hitfile)

        nt.assert_equal(hits.data, checkhits.data)

        return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import numpy as np

import healpy as hp

from ..timing import function_timer

from .todmap_math import dipole

from ..op import Operator

from ..utils import Environment


class OpSimDipole(Operator):
    """
    Operator which generates dipole signal for detectors.

    This uses the detector pointing, the telescope velocity vectors, and
    the solar system motion with respect to the CMB rest frame to compute
    the observed CMB dipole signal.  The dipole timestream is either added
    (default) or subtracted from a cache object.

    The telescope velocity and detector quaternions are assumed to be in
    the same coordinate system.

    Args:
        mode (str): this determines what components of the telescope motion
            are included in the observed dipole.  Valid options are 'solar'
            for just the solar system motion, 'orbital' for just the motion
            of the telescope with respect to the solarsystem barycenter, and
            'total' which is the sum of both (and the default).
        coord (str): coordinate system of detector pointing.  Valid values
            are 'C' for equatorial, 'E' for ecliptic and 'G' for galactic.
        subtract (bool): if True, subtract timestream from cache object,
            otherwise add it (default).
        out (str): accumulate data to the cache with name <out>_<detector>.
            If the named cache objects do not exist, then they are created.
        cmb (float): CMB monopole in Kelvin.  Default value from Fixsen
            2009 (see arXiv:0911.1955)
        solar_speed (float): the amplitude of the solarsystem barycenter
            velocity with respect to the CMB in Km/s.  The default value is
            based on http://arxiv.org/abs/0803.0732.
        solar_gal_lat (float): the latitude in degrees in galactic
            coordinates for the direction of motion of the solarsystem with
            respect to the CMB rest frame.
        solar_gal_lon (float): the longitude in degrees in galactic
            coordinates for the direction of motion of the solarsystem with
            respect to the CMB rest frame.
        freq (float): optional observing frequency in Hz (not GHz).
    """

    def __init__(
        self,
        mode="total",
        coord="C",
        subtract=False,
        out="dipole",
        cmb=2.72548,
        solar_speed=369.0,
        solar_gal_lat=48.26,
        solar_gal_lon=263.99,
        freq=0,
        keep_quats=False,
        keep_vel=False,
        flag_mask=255,
        common_flag_mask=255,
    ):
        self._mode = mode
        self._coord = coord
        self._subtract = subtract
        self._out = out
        self._cmb = cmb
        self._freq = freq
        self._solar_speed = solar_speed
        self._solar_gal_theta = np.deg2rad(90.0 - solar_gal_lat)
        self._solar_gal_phi = np.deg2rad(solar_gal_lon)
        self._keep_quats = keep_quats
        self._keep_vel = keep_vel
        self._flag_mask = flag_mask
        self._common_flag_mask = common_flag_mask

        projected = self._solar_speed * np.sin(self._solar_gal_theta)
        z = self._solar_speed * np.cos(self._solar_gal_theta)
        x = projected * np.cos(self._solar_gal_phi)
        y = projected * np.sin(self._solar_gal_phi)
        self._solar_gal_vel = np.array([x, y, z])

        # rotate solar system velocity to desired coordinate frame

        if self._coord == "G":
            self._solar_vel = self._solar_gal_vel
        else:
            rotmat = hp.rotator.Rotator(coord=["G", self._coord]).mat
            self._solar_vel = np.ravel(np.dot(rotmat, self._solar_gal_vel))

        super().__init__()

    @function_timer
    def exec(self, data):
        """Create the timestreams.

        This loops over all observations and detectors and uses the pointing,
        the telescope motion, and the solar system motion to compute the
        observed dipole.

        Args:
            data (toast.Data): The distributed data.

        """
        env = Environment.get()
        comm = data.comm
        # the global communicator
        cworld = comm.comm_world
        # the communicator within the group
        cgroup = comm.comm_group
        # the communicator with all processes with
        # the same rank within their group
        crank = comm.comm_rank

        nullquat = np.array([0, 0, 0, 1], dtype=np.float64)

        for obs in data.obs:
            tod = obs["tod"]

            offset, nsamp = tod.local_samples

            vel = None
            sol = None

            if (self._mode == "solar") or (self._mode == "total"):
                sol = self._solar_vel

            if (self._mode == "orbital") or (self._mode == "total"):
                if self._keep_vel:
                    vel = tod.local_velocity()
                else:
                    vel = tod.read_velocity()

            common = tod.local_common_flags() & self._common_flag_mask

            for det in tod.local_dets:

                flags = tod.local_flags(det) & self._flag_mask
                totflags = flags | common
                del flags

                pdata = None
                if self._keep_quats:
                    # We are keeping the detector quaternions, so cache
                    # them now for the full sample range.
                    pdata = tod.local_pointing(det)

                # Set up output cache
                cachename = "{}_{}".format(self._out, det)
                if not tod.cache.exists(cachename):
                    tod.cache.create(cachename, np.float64, (nsamp,))
                ref = tod.cache.reference(cachename)

                buf_off = 0
                buf_n = env.tod_buffer_length()
                while buf_off < nsamp:
                    if buf_off + buf_n > nsamp:
                        buf_n = nsamp - buf_off
                    bslice = slice(buf_off, buf_off + buf_n)

                    detp = None
                    if pdata is None:
                        # Read and discard
                        detp = tod.read_pntg(detector=det, local_start=buf_off, n=buf_n)
                    else:
                        # Use cached version
                        detp = pdata[bslice, :]

                    # Make sure that flagged pointing is well defined
                    detp[(totflags[bslice] != 0), :] = nullquat

                    vslice = None
                    if vel is not None:
                        vslice = vel[bslice]

                    dipoletod = dipole(
                        detp, vel=vslice, solar=sol, cmb=self._cmb, freq=self._freq
                    )
                    if self._subtract:
                        ref[bslice] -= dipoletod
                    else:
                        ref[bslice] += dipoletod
                    buf_off += buf_n

                del pdata
                del ref

            del vel
            del common

        return
